# -*- coding: utf-8 -*-
#   Copyright 2021 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Tests for projectq.backends._ionq._ionq.py."""

import math
from unittest import mock

import pytest

from projectq import MainEngine
from projectq.backends._ionq import _ionq, _ionq_http_client
from projectq.backends._ionq._ionq_exc import (
    InvalidCommandError,
    MidCircuitMeasurementError,
)
from projectq.backends._ionq._ionq_mapper import BoundedQubitMapper
from projectq.cengines import DummyEngine
from projectq.ops import (
    CNOT,
    All,
    Allocate,
    Barrier,
    Command,
    Deallocate,
    Entangle,
    H,
    Measure,
    Ph,
    R,
    Rx,
    Rxx,
    Ry,
    Rz,
    S,
    Sdag,
    SqrtX,
    T,
    Tdag,
    Toffoli,
    X,
    Y,
    Z,
)
from projectq.types import WeakQubitRef


@pytest.fixture(scope='function')
def mapper_factory():
    def _factory(n=4):
        return BoundedQubitMapper(n)

    return _factory


# Prevent any requests from making it out.
@pytest.fixture(autouse=True)
def no_requests(monkeypatch):
    monkeypatch.delattr("requests.sessions.Session.request")


@pytest.mark.parametrize(
    "single_qubit_gate, is_available",
    [
        (X, True),
        (Y, True),
        (Z, True),
        (H, True),
        (T, True),
        (Tdag, True),
        (S, True),
        (Sdag, True),
        (Allocate, True),
        (Deallocate, True),
        (SqrtX, True),
        (Measure, True),
        (Rx(0.5), True),
        (Ry(0.5), True),
        (Rz(0.5), True),
        (R(0.5), False),
        (Barrier, True),
        (Entangle, False),
    ],
)
def test_ionq_backend_is_available(single_qubit_gate, is_available):
    eng = MainEngine(backend=DummyEngine(), engine_list=[DummyEngine()])
    qubit1 = eng.allocate_qubit()
    ionq_backend = _ionq.IonQBackend()
    cmd = Command(eng, single_qubit_gate, (qubit1,))
    assert ionq_backend.is_available(cmd) is is_available


# IonQ supports up to 7 control qubits.
@pytest.mark.parametrize(
    "num_ctrl_qubits, is_available",
    [
        (0, True),
        (1, True),
        (2, True),
        (3, True),
        (4, True),
        (5, True),
        (6, True),
        (7, True),
        (8, False),
    ],
)
def test_ionq_backend_is_available_control_not(num_ctrl_qubits, is_available):
    eng = MainEngine(backend=DummyEngine(), engine_list=[DummyEngine()])
    qubit1 = eng.allocate_qubit()
    qureg = eng.allocate_qureg(num_ctrl_qubits)
    ionq_backend = _ionq.IonQBackend()
    cmd = Command(eng, X, (qubit1,), controls=qureg)
    assert ionq_backend.is_available(cmd) is is_available


def test_ionq_backend_is_available_negative_control():
    backend = _ionq.IonQBackend()

    qb0 = WeakQubitRef(engine=None, idx=0)
    qb1 = WeakQubitRef(engine=None, idx=1)
    qb2 = WeakQubitRef(engine=None, idx=2)

    assert backend.is_available(Command(None, X, qubits=([qb0],), controls=[qb1]))
    assert backend.is_available(Command(None, X, qubits=([qb0],), controls=[qb1], control_state='1'))
    assert not backend.is_available(Command(None, X, qubits=([qb0],), controls=[qb1], control_state='0'))

    assert backend.is_available(Command(None, X, qubits=([qb0],), controls=[qb1, qb2]))
    assert backend.is_available(Command(None, X, qubits=([qb0],), controls=[qb1, qb2], control_state='11'))
    assert not backend.is_available(Command(None, X, qubits=([qb0],), controls=[qb1, qb2], control_state='01'))


def test_ionq_backend_init():
    """Test initialized backend has an empty circuit"""
    backend = _ionq.IonQBackend(verbose=True, use_hardware=True)
    assert hasattr(backend, '_circuit')
    circuit = getattr(backend, '_circuit')
    assert isinstance(circuit, list)
    assert len(circuit) == 0


def test_ionq_empty_circuit():
    """Test that empty circuits are still flushable."""
    backend = _ionq.IonQBackend(verbose=True)
    eng = MainEngine(backend=backend)
    eng.flush()


def test_ionq_no_circuit_executed():
    """Test that one can't retrieve probabilities if no circuit was run."""
    backend = _ionq.IonQBackend(verbose=True)
    eng = MainEngine(backend=backend)
    # no circuit has been executed -> raises exception
    with pytest.raises(RuntimeError):
        backend.get_probabilities([])
    eng.flush()


def test_ionq_get_probability(monkeypatch, mapper_factory):
    """Test a shortcut for getting a specific state's probability"""

    def mock_retrieve(*args, **kwargs):
        return {
            'nq': 3,
            'shots': 10,
            'output_probs': {'3': 0.4, '0': 0.6},
            'meas_mapped': [0, 1],
            'meas_qubit_ids': [1, 2],
        }

    monkeypatch.setattr(_ionq_http_client, "retrieve", mock_retrieve)
    backend = _ionq.IonQBackend(
        retrieve_execution="a3877d18-314f-46c9-86e7-316bc4dbe968",
        verbose=True,
    )
    eng = MainEngine(backend=backend, engine_list=[mapper_factory()])

    unused_qubit = eng.allocate_qubit()  # noqa: F841
    qureg = eng.allocate_qureg(2)
    # entangle the qureg
    Ry(math.pi / 2) | qureg[0]
    Rx(math.pi / 2) | qureg[0]
    Rx(math.pi / 2) | qureg[0]
    Ry(math.pi / 2) | qureg[0]
    Rxx(math.pi / 2) | (qureg[0], qureg[1])
    Rx(7 * math.pi / 2) | qureg[0]
    Ry(7 * math.pi / 2) | qureg[0]
    Rx(7 * math.pi / 2) | qureg[1]

    # measure; should be all-0 or all-1
    All(Measure) | qureg
    # run the circuit
    eng.flush()
    assert eng.backend.get_probability('11', qureg) == pytest.approx(0.4)
    assert eng.backend.get_probability('00', qureg) == pytest.approx(0.6)

    with pytest.raises(ValueError) as excinfo:
        eng.backend.get_probability('111', qureg)
    assert str(excinfo.value) == 'Desired state and register must be the same length!'


def test_ionq_get_probabilities(monkeypatch, mapper_factory):
    """Test a shortcut for getting a specific state's probability"""

    def mock_retrieve(*args, **kwargs):
        return {
            'nq': 3,
            'shots': 10,
            'output_probs': {'1': 0.4, '0': 0.6},
            'meas_mapped': [1],
            'meas_qubit_ids': [1],
        }

    monkeypatch.setattr(_ionq_http_client, "retrieve", mock_retrieve)
    backend = _ionq.IonQBackend(
        retrieve_execution="a3877d18-314f-46c9-86e7-316bc4dbe968",
        verbose=True,
    )
    eng = MainEngine(backend=backend, engine_list=[mapper_factory()])
    qureg = eng.allocate_qureg(2)
    q0, q1 = qureg
    H | q0
    CNOT | (q0, q1)
    Measure | q1
    # run the circuit
    eng.flush()
    assert eng.backend.get_probability('01', qureg) == pytest.approx(0.4)
    assert eng.backend.get_probability('00', qureg) == pytest.approx(0.6)
    assert eng.backend.get_probability('1', [qureg[1]]) == pytest.approx(0.4)
    assert eng.backend.get_probability('0', [qureg[1]]) == pytest.approx(0.6)


def test_ionq_invalid_command():
    """Test that this backend raises out with invalid commands."""

    # Ph gate is not a valid gate
    qb = WeakQubitRef(None, 1)
    cmd = Command(None, gate=Ph(math.pi), qubits=[(qb,)])
    backend = _ionq.IonQBackend(verbose=True)
    with pytest.raises(InvalidCommandError):
        backend.receive([cmd])


def test_ionq_sent_error(monkeypatch, mapper_factory):
    """Test that errors on "send" will raise back out."""
    # patch send
    type_error = TypeError()
    mock_send = mock.MagicMock(side_effect=type_error)
    monkeypatch.setattr(_ionq_http_client, "send", mock_send)

    backend = _ionq.IonQBackend()
    eng = MainEngine(
        backend=backend,
        engine_list=[mapper_factory()],
        verbose=True,
    )
    qubit = eng.allocate_qubit()
    Rx(0.5) | qubit
    with pytest.raises(Exception) as excinfo:
        qubit[0].__del__()
        eng.flush()

    # verbose=True on the engine re-raises errors instead of compacting them.
    assert type_error is excinfo.value

    # atexit sends another FlushGate, therefore we remove the backend:
    dummy = DummyEngine()
    dummy.is_last_engine = True
    eng.next_engine = dummy


def test_ionq_send_nonetype_response_error(monkeypatch, mapper_factory):
    """Test that no return value from "send" will raise a runtime error."""
    # patch send
    mock_send = mock.MagicMock(return_value=None)
    monkeypatch.setattr(_ionq_http_client, "send", mock_send)

    backend = _ionq.IonQBackend()
    eng = MainEngine(
        backend=backend,
        engine_list=[mapper_factory()],
        verbose=True,
    )
    qubit = eng.allocate_qubit()
    Rx(0.5) | qubit
    with pytest.raises(RuntimeError) as excinfo:
        eng.flush()

    # verbose=True on the engine re-raises errors instead of compacting them.
    assert str(excinfo.value) == "Failed to submit job to the server!"

    # atexit sends another FlushGate, therefore we remove the backend:
    dummy = DummyEngine()
    dummy.is_last_engine = True
    eng.next_engine = dummy


def test_ionq_retrieve(monkeypatch, mapper_factory):
    """Test that initializing a backend with a jobid will fetch that job's results to use as its own"""

    def mock_retrieve(*args, **kwargs):
        return {
            'nq': 3,
            'shots': 10,
            'output_probs': {'3': 0.4, '0': 0.6},
            'meas_mapped': [0, 1],
            'meas_qubit_ids': [1, 2],
        }

    monkeypatch.setattr(_ionq_http_client, "retrieve", mock_retrieve)
    backend = _ionq.IonQBackend(
        retrieve_execution="a3877d18-314f-46c9-86e7-316bc4dbe968",
        verbose=True,
    )
    eng = MainEngine(backend=backend, engine_list=[mapper_factory()])

    unused_qubit = eng.allocate_qubit()
    qureg = eng.allocate_qureg(2)
    # entangle the qureg
    Ry(math.pi / 2) | qureg[0]
    Rx(math.pi / 2) | qureg[0]
    Rx(math.pi / 2) | qureg[0]
    Ry(math.pi / 2) | qureg[0]
    Rxx(math.pi / 2) | (qureg[0], qureg[1])
    Rx(7 * math.pi / 2) | qureg[0]
    Ry(7 * math.pi / 2) | qureg[0]
    Rx(7 * math.pi / 2) | qureg[1]
    del unused_qubit
    # measure; should be all-0 or all-1
    All(Measure) | qureg
    # run the circuit
    eng.flush()
    prob_dict = eng.backend.get_probabilities([qureg[0], qureg[1]])
    assert prob_dict['11'] == pytest.approx(0.4)
    assert prob_dict['00'] == pytest.approx(0.6)

    # Unknown qubit
    invalid_qubit = [WeakQubitRef(eng, 10)]
    probs = eng.backend.get_probabilities(invalid_qubit)
    assert {'0': 1} == probs


def test_ionq_retrieve_nonetype_response_error(monkeypatch, mapper_factory):
    """Test that initializing a backend with a jobid will fetch that job's results to use as its own"""

    def mock_retrieve(*args, **kwargs):
        return None

    monkeypatch.setattr(_ionq_http_client, "retrieve", mock_retrieve)
    backend = _ionq.IonQBackend(
        retrieve_execution="a3877d18-314f-46c9-86e7-316bc4dbe968",
        verbose=True,
    )
    eng = MainEngine(
        backend=backend,
        engine_list=[mapper_factory()],
        verbose=True,
    )

    unused_qubit = eng.allocate_qubit()
    qureg = eng.allocate_qureg(2)
    # entangle the qureg
    Ry(math.pi / 2) | qureg[0]
    Rx(math.pi / 2) | qureg[0]
    Rx(math.pi / 2) | qureg[0]
    Ry(math.pi / 2) | qureg[0]
    Rxx(math.pi / 2) | (qureg[0], qureg[1])
    Rx(7 * math.pi / 2) | qureg[0]
    Ry(7 * math.pi / 2) | qureg[0]
    Rx(7 * math.pi / 2) | qureg[1]
    del unused_qubit
    # measure; should be all-0 or all-1
    All(Measure) | qureg
    # run the circuit
    with pytest.raises(RuntimeError) as excinfo:
        eng.flush()

    exc = excinfo.value
    expected_err = "Failed to retrieve job with id: 'a3877d18-314f-46c9-86e7-316bc4dbe968'!"
    assert str(exc) == expected_err


def test_ionq_backend_functional_test(monkeypatch, mapper_factory):
    """Test that the backend can handle a valid circuit with valid results."""
    expected = {
        'nq': 3,
        'shots': 10,
        'meas_mapped': [1, 2],
        'meas_qubit_ids': [1, 2],
        'circuit': [
            {'gate': 'ry', 'rotation': 0.5, 'targets': [1]},
            {'gate': 'rx', 'rotation': 0.5, 'targets': [1]},
            {'gate': 'rx', 'rotation': 0.5, 'targets': [1]},
            {'gate': 'ry', 'rotation': 0.5, 'targets': [1]},
            {'gate': 'xx', 'rotation': 0.5, 'targets': [1, 2]},
            {'gate': 'rx', 'rotation': 3.5, 'targets': [1]},
            {'gate': 'ry', 'rotation': 3.5, 'targets': [1]},
            {'gate': 'rx', 'rotation': 3.5, 'targets': [2]},
        ],
    }

    def mock_send(*args, **kwargs):
        assert args[0] == expected
        return {
            'nq': 3,
            'shots': 10,
            'output_probs': {'3': 0.4, '0': 0.6},
            'meas_mapped': [1, 2],
            'meas_qubit_ids': [1, 2],
        }

    monkeypatch.setattr(_ionq_http_client, "send", mock_send)
    backend = _ionq.IonQBackend(verbose=True, num_runs=10)
    eng = MainEngine(
        backend=backend,
        engine_list=[mapper_factory()],
        verbose=True,
    )
    unused_qubit = eng.allocate_qubit()  # noqa: F841
    qureg = eng.allocate_qureg(2)

    # entangle the qureg
    Ry(0.5) | qureg[0]
    Rx(0.5) | qureg[0]
    Rx(0.5) | qureg[0]
    Ry(0.5) | qureg[0]
    Rxx(0.5) | (qureg[0], qureg[1])
    Rx(3.5) | qureg[0]
    Ry(3.5) | qureg[0]
    Rx(3.5) | qureg[1]
    All(Barrier) | qureg
    # measure; should be all-0 or all-1
    All(Measure) | qureg
    # run the circuit
    eng.flush()
    prob_dict = eng.backend.get_probabilities([qureg[0], qureg[1]])
    assert prob_dict['11'] == pytest.approx(0.4)
    assert prob_dict['00'] == pytest.approx(0.6)


def test_ionq_backend_functional_aliases_test(monkeypatch, mapper_factory):
    """Test that sub-classed or aliased gates are handled correctly."""
    # using alias gates, for coverage
    expected = {
        'nq': 4,
        'shots': 10,
        'meas_mapped': [2, 3],
        'meas_qubit_ids': [2, 3],
        'circuit': [
            {'gate': 'x', 'targets': [0]},
            {'gate': 'x', 'targets': [1]},
            {'controls': [0], 'gate': 'x', 'targets': [2]},
            {'controls': [1], 'gate': 'x', 'targets': [2]},
            {'controls': [0, 1], 'gate': 'x', 'targets': [3]},
            {'gate': 's', 'targets': [2]},
            {'gate': 'si', 'targets': [3]},
        ],
    }

    def mock_send(*args, **kwargs):
        assert args[0] == expected
        return {
            'nq': 4,
            'shots': 10,
            'output_probs': {'1': 0.9},
            'meas_mapped': [2, 3],
        }

    monkeypatch.setattr(_ionq_http_client, "send", mock_send)
    backend = _ionq.IonQBackend(verbose=True, num_runs=10)
    eng = MainEngine(
        backend=backend,
        engine_list=[mapper_factory(9)],
        verbose=True,
    )
    # Do some stuff with a circuit. Get weird with it.
    circuit = eng.allocate_qureg(4)
    qubit1, qubit2, qubit3, qubit4 = circuit
    All(X) | [qubit1, qubit2]
    CNOT | (qubit1, qubit3)
    CNOT | (qubit2, qubit3)
    Toffoli | (qubit1, qubit2, qubit4)
    Barrier | circuit
    S | qubit3
    Sdag | qubit4
    All(Measure) | [qubit3, qubit4]

    # run the circuit
    eng.flush()
    prob_dict = eng.backend.get_probabilities([qubit3, qubit4])
    assert prob_dict['10'] == pytest.approx(0.9)


def test_ionq_no_midcircuit_measurement(monkeypatch, mapper_factory):
    """Test that attempts to measure mid-circuit raise exceptions."""

    def mock_send(*args, **kwargs):
        return {
            'nq': 1,
            'shots': 10,
            'output_probs': {'0': 0.4, '1': 0.6},
        }

    monkeypatch.setattr(_ionq_http_client, "send", mock_send)

    # Create a backend to use with an engine.
    backend = _ionq.IonQBackend(verbose=True, num_runs=10)
    eng = MainEngine(
        backend=backend,
        engine_list=[mapper_factory()],
        verbose=True,
    )
    qubit = eng.allocate_qubit()
    X | qubit
    Measure | qubit
    with pytest.raises(MidCircuitMeasurementError):
        X | qubit

    # atexit sends another FlushGate, therefore we remove the backend:
    dummy = DummyEngine()
    dummy.is_last_engine = True
    eng.active_qubits = []
    eng.next_engine = dummy



# -*- coding: utf-8 -*-
#   Copyright 2021 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
# pylint: skip-file

"""Example of a basic 'half-adder' circuit using an IonQBackend"""

import getpass
import random

import matplotlib.pyplot as plt

import projectq.setups.default
import projectq.setups.ionq
from projectq import MainEngine
from projectq.backends import IonQBackend
from projectq.libs.hist import histogram
from projectq.ops import CNOT, All, Barrier, Measure, Toffoli, X


def run_half_adder(eng):
    # allocate the quantum register to entangle
    circuit = eng.allocate_qureg(4)
    qubit1, qubit2, qubit3, qubit4 = circuit
    result_qubits = [qubit3, qubit4]

    # X gates on the first two qubits
    All(X) | [qubit1, qubit2]

    # Barrier
    Barrier | circuit

    # Cx gates
    CNOT | (qubit1, qubit3)
    CNOT | (qubit2, qubit3)

    # CCNOT
    Toffoli | (qubit1, qubit2, qubit4)

    # Barrier
    Barrier | circuit

    # Measure result qubits
    All(Measure) | result_qubits

    # Flush the circuit (this submits a job to the IonQ API)
    eng.flush()

    # Show the histogram
    histogram(eng.backend, result_qubits)
    plt.show()

    # return a random answer from our results
    probabilities = eng.backend.get_probabilities(result_qubits)
    random_answer = random.choice(list(probabilities.keys()))
    return [int(s) for s in random_answer]


if __name__ == '__main__':
    token = None
    device = None
    if token is None:
        token = getpass.getpass(prompt='IonQ apiKey > ')
    if device is None:
        device = input('IonQ device > ')

    backend = IonQBackend(
        use_hardware=True,
        token=token,
        num_runs=200,
        verbose=True,
        device=device,
    )
    engine_list = projectq.setups.ionq.get_engine_list(
        token=token,
        device=device,
    )
    engine = MainEngine(backend, engine_list)
    # run the circuit and print the result
    print(run_half_adder(engine))



# -*- coding: utf-8 -*-
#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

""" Back-end to run quantum program on IBM QE cloud platform"""


# helpers to run the jsonified gate sequence on ibm quantum experience server
# api documentation does not exist and has to be deduced from the qiskit code
# source at: https://github.com/Qiskit/qiskit-ibmq-provider

import getpass
import time
import signal
import uuid

import requests
from requests.compat import urljoin
from requests import Session

_AUTH_API_URL = 'https://auth.quantum-computing.ibm.com/api/users/loginWithToken'
_API_URL = 'https://api.quantum-computing.ibm.com/api/'

# TODO: call to get the API version automatically
CLIENT_APPLICATION = 'ibmqprovider/0.4.4'


class IBMQ(Session):
    """
    Manage a session between ProjectQ and the IBMQ web API.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.backends = dict()
        self.timeout = 5.0

    def get_list_devices(self, verbose=False):
        """
        Get the list of available IBM backends with their properties

        Args:
            verbose (bool): print the returned dictionnary if True

        Returns:
            (dict) backends dictionary by name device, containing the qubit
                    size 'nq', the coupling map 'coupling_map' as well as the
                    device version 'version'
        """
        list_device_url = 'Network/ibm-q/Groups/open/Projects/main/devices/v/1'
        argument = {'allow_redirects': True, 'timeout': (self.timeout, None)}
        request = super().get(urljoin(_API_URL, list_device_url), **argument)
        request.raise_for_status()
        r_json = request.json()
        self.backends = dict()
        for obj in r_json:
            self.backends[obj['backend_name']] = {
                'nq': obj['n_qubits'],
                'coupling_map': obj['coupling_map'],
                'version': obj['backend_version'],
            }

        if verbose:
            print('- List of IBMQ devices available:')
            print(self.backends)
        return self.backends

    def is_online(self, device):
        """
        Check if the device is in the list of available IBM backends.

        Args:
            device (str): name of the device to check

        Returns:
            (bool) True if device is available, False otherwise
        """
        return device in self.backends

    def can_run_experiment(self, info, device):
        """
        Check if the device is big enough to run the code.

        Args:
            info (dict): dictionary sent by the backend containing the code to
                run
            device (str): name of the ibm device to use

        Returns:
            (tuple): (bool) True if device is big enough, False otherwise
                     (int) maximum number of qubit available on the device
                     (int) number of qubit needed for the circuit

        """
        nb_qubit_max = self.backends[device]['nq']
        nb_qubit_needed = info['nq']
        return nb_qubit_needed <= nb_qubit_max, nb_qubit_max, nb_qubit_needed

    def authenticate(self, token=None):
        """
        Args:
            token (str): IBM quantum experience user API token.
        """
        if token is None:
            token = getpass.getpass(prompt="IBM QE token > ")
        if len(token) == 0:
            raise Exception('Error with the IBM QE token')
        self.headers.update({'X-Qx-Client-Application': CLIENT_APPLICATION})
        args = {
            'data': None,
            'json': {'apiToken': token},
            'timeout': (self.timeout, None),
        }
        request = super().post(_AUTH_API_URL, **args)
        request.raise_for_status()
        r_json = request.json()
        self.params.update({'access_token': r_json['id']})

    def run(self, info, device):  # pylint: disable=too-many-locals
        """
        Run the quantum code to the IBMQ machine.
        Update since September 2020: only protocol available is what they call
        'object storage' where a job request via the POST method gets in
        return a url link to which send the json data. A final http validates
        the data communication.

        Args:
            info (dict): dictionary sent by the backend containing the code to
                run
            device (str): name of the ibm device to use

        Returns:
            (tuple): (str) Execution Id

        """

        # STEP1: Obtain most of the URLs for handling communication with
        #        quantum device
        json_step1 = {
            'data': None,
            'json': {
                'backend': {'name': device},
                'allowObjectStorage': True,
                'shareLevel': 'none',
            },
            'timeout': (self.timeout, None),
        }
        request = super().post(
            urljoin(_API_URL, 'Network/ibm-q/Groups/open/Projects/main/Jobs'),
            **json_step1,
        )
        request.raise_for_status()
        r_json = request.json()
        upload_url = r_json['objectStorageInfo']['uploadUrl']
        execution_id = r_json['id']

        # STEP2: WE UPLOAD THE CIRCUIT DATA
        n_classical_reg = info['nq']
        # hack: easier to restrict labels to measured qubits
        n_qubits = n_classical_reg  # self.backends[device]['nq']
        instructions = info['json']
        maxcredit = info['maxCredits']
        c_label = [["c", i] for i in range(n_classical_reg)]
        q_label = [["q", i] for i in range(n_qubits)]

        # hack: the data value in the json quantum code is a string
        instruction_str = str(instructions).replace('\'', '\"')
        data = '{"qobj_id": "' + str(uuid.uuid4()) + '", '
        data += '"header": {"backend_name": "' + device + '", '
        data += '"backend_version": "' + self.backends[device]['version'] + '"}, '
        data += '"config": {"shots": ' + str(info['shots']) + ', '
        data += '"max_credits": ' + str(maxcredit) + ', "memory": false, '
        data += '"parameter_binds": [], "memory_slots": ' + str(n_classical_reg)
        data += ', "n_qubits": ' + str(n_qubits) + '}, "schema_version": "1.2.0", '
        data += '"type": "QASM", "experiments": [{"config": '
        data += '{"n_qubits": ' + str(n_qubits) + ', '
        data += '"memory_slots": ' + str(n_classical_reg) + '}, '
        data += '"header": {"qubit_labels": ' + str(q_label).replace('\'', '\"') + ', '
        data += '"n_qubits": ' + str(n_classical_reg) + ', '
        data += '"qreg_sizes": [["q", ' + str(n_qubits) + ']], '
        data += '"clbit_labels": ' + str(c_label).replace('\'', '\"') + ', '
        data += '"memory_slots": ' + str(n_classical_reg) + ', '
        data += '"creg_sizes": [["c", ' + str(n_classical_reg) + ']], '
        data += '"name": "circuit0", "global_phase": 0}, "instructions": ' + instruction_str + '}]}'

        json_step2 = {
            'data': data,
            'params': {'access_token': None},
            'timeout': (5.0, None),
        }
        request = super().put(upload_url, **json_step2)
        request.raise_for_status()

        # STEP3: CONFIRM UPLOAD
        json_step3 = {'data': None, 'json': None, 'timeout': (self.timeout, None)}

        upload_data_url = urljoin(
            _API_URL,
            'Network/ibm-q/Groups/open/Projects/main/Jobs/' + str(execution_id) + '/jobDataUploaded',
        )
        request = super().post(upload_data_url, **json_step3)
        request.raise_for_status()

        return execution_id

    def get_result(
        self, device, execution_id, num_retries=3000, interval=1, verbose=False
    ):  # pylint: disable=too-many-arguments,too-many-locals
        """
        Get the result of an execution
        """

        job_status_url = 'Network/ibm-q/Groups/open/Projects/main/Jobs/' + execution_id

        if verbose:
            print("Waiting for results. [Job ID: {}]".format(execution_id))

        original_sigint_handler = signal.getsignal(signal.SIGINT)

        def _handle_sigint_during_get_result(*_):  # pragma: no cover
            raise Exception("Interrupted. The ID of your submitted job is {}.".format(execution_id))

        try:
            signal.signal(signal.SIGINT, _handle_sigint_during_get_result)
            for retries in range(num_retries):

                # STEP5: WAIT FOR THE JOB TO BE RUN
                json_step5 = {'allow_redirects': True, 'timeout': (self.timeout, None)}
                request = super().get(urljoin(_API_URL, job_status_url), **json_step5)
                request.raise_for_status()
                r_json = request.json()
                acceptable_status = ['VALIDATING', 'VALIDATED', 'RUNNING']
                if r_json['status'] == 'COMPLETED':
                    # STEP6: Get the endpoint to get the result
                    json_step6 = {
                        'allow_redirects': True,
                        'timeout': (self.timeout, None),
                    }
                    request = super().get(
                        urljoin(_API_URL, job_status_url + '/resultDownloadUrl'),
                        **json_step6,
                    )
                    request.raise_for_status()
                    r_json = request.json()

                    # STEP7: Get the result
                    json_step7 = {
                        'allow_redirects': True,
                        'params': {'access_token': None},
                        'timeout': (self.timeout, None),
                    }
                    request = super().get(r_json['url'], **json_step7)
                    r_json = request.json()
                    result = r_json['results'][0]

                    # STEP8: Confirm the data was downloaded
                    json_step8 = {'data': None, 'json': None, 'timeout': (5.0, None)}
                    request = super().post(
                        urljoin(_API_URL, job_status_url + '/resultDownloaded'),
                        **json_step8,
                    )
                    r_json = request.json()
                    return result

                # Note: if stays stuck if 'Validating' mode, then sthg went
                #       wrong in step 3
                if r_json['status'] not in acceptable_status:
                    raise Exception("Error while running the code. Last status: {}.".format(r_json['status']))
                time.sleep(interval)
                if self.is_online(device) and retries % 60 == 0:
                    self.get_list_devices()
                    if not self.is_online(device):
                        raise DeviceOfflineError(
                            "Device went offline. The ID of your submitted job is {}.".format(execution_id)
                        )

        finally:
            if original_sigint_handler is not None:
                signal.signal(signal.SIGINT, original_sigint_handler)

        raise Exception("Timeout. The ID of your submitted job is {}.".format(execution_id))


class DeviceTooSmall(Exception):
    """Exception raised if the device is too small to run the circuit"""


class DeviceOfflineError(Exception):
    """Exception raised if a selected device is currently offline"""


def show_devices(token=None, verbose=False):
    """
    Access the list of available devices and their properties (ex: for setup
    configuration)

    Args:
        token (str): IBM quantum experience user API token.
        verbose (bool): If True, additional information is printed

    Returns:
        (list) list of available devices and their properties
    """
    ibmq_session = IBMQ()
    ibmq_session.authenticate(token=token)
    return ibmq_session.get_list_devices(verbose=verbose)


def retrieve(device, token, jobid, num_retries=3000, interval=1, verbose=False):  # pylint: disable=too-many-arguments
    """
    Retrieves a previously run job by its ID.

    Args:
        device (str): Device on which the code was run / is running.
        token (str): IBM quantum experience user API token.
        jobid (str): Id of the job to retrieve

    Returns:
        (dict) result form the IBMQ server
    """
    ibmq_session = IBMQ()
    ibmq_session.authenticate(token)
    ibmq_session.get_list_devices(verbose)
    res = ibmq_session.get_result(device, jobid, num_retries=num_retries, interval=interval, verbose=verbose)
    return res


def send(
    info,
    device='ibmq_qasm_simulator',
    token=None,
    shots=None,
    num_retries=3000,
    interval=1,
    verbose=False,
):  # pylint: disable=too-many-arguments
    """
    Sends QASM through the IBM API and runs the quantum circuit.

    Args:
        info(dict): Contains representation of the circuit to run.
        device (str): name of the ibm device. Simulator chosen by default
        token (str): IBM quantum experience user API token.
        shots (int): Number of runs of the same circuit to collect statistics.
        verbose (bool): If True, additional information is printed, such as
            measurement statistics. Otherwise, the backend simply registers
            one measurement result (same behavior as the projectq Simulator).

    Returns:
        (dict) result form the IBMQ server

    """
    try:
        ibmq_session = IBMQ()
        # Shots argument deprecated, as already
        if shots is not None:
            info['shots'] = shots
        if verbose:
            print("- Authenticating...")
            if token is not None:
                print('user API token: ' + token)
        ibmq_session.authenticate(token)

        # check if the device is online
        ibmq_session.get_list_devices(verbose)
        online = ibmq_session.is_online(device)
        if not online:
            print("The device is offline (for maintenance?). Use the simulator instead or try again later.")
            raise DeviceOfflineError("Device is offline.")

        # check if the device has enough qubit to run the code
        runnable, qmax, qneeded = ibmq_session.can_run_experiment(info, device)
        if not runnable:
            print(
                (
                    "The device is too small ({} qubits available) for the code "
                    + "requested({} qubits needed) Try to look for another "
                    + "device with more qubits"
                ).format(qmax, qneeded)
            )
            raise DeviceTooSmall("Device is too small.")
        if verbose:
            print("- Running code: {}".format(info))
        execution_id = ibmq_session.run(info, device)
        if verbose:
            print("- Waiting for results...")
        res = ibmq_session.get_result(
            device,
            execution_id,
            num_retries=num_retries,
            interval=interval,
            verbose=verbose,
        )
        if verbose:
            print("- Done.")
        return res
    except requests.exceptions.HTTPError as err:
        print("- There was an error running your code:")
        print(err)
    except requests.exceptions.RequestException as err:
        print("- Looks like something is wrong with server:")
        print(err)
    except KeyError as err:
        print("- Failed to parse response:")
        print(err)
    return None



# -*- coding: utf-8 -*-
#   Copyright 2021 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
""" Back-end to run quantum program on AWS Braket provided devices."""

import random
import json

from projectq.cengines import BasicEngine
from projectq.meta import get_control_count, LogicalQubitIDTag, has_negative_control
from projectq.types import WeakQubitRef
from projectq.ops import (
    R,
    SwapGate,
    HGate,
    Rx,
    Ry,
    Rz,
    SGate,
    Sdag,
    TGate,
    Tdag,
    XGate,
    YGate,
    ZGate,
    SqrtXGate,
    Measure,
    Allocate,
    Deallocate,
    Barrier,
    FlushGate,
    DaggeredGate,
)

# TODO: Add MatrixGate to cover the unitary operation in the SV1 simulator

from ._awsbraket_boto3_client import send, retrieve


class AWSBraketBackend(BasicEngine):  # pylint: disable=too-many-instance-attributes
    """
    The AWS Braket Backend class, which stores the circuit, transforms it to Braket compatible, and sends the circuit
    through the Boto3 and Amazon Braket SDK.
    """

    def __init__(
        self,
        use_hardware=False,
        num_runs=1000,
        verbose=False,
        credentials=None,
        s3_folder=None,
        device='Aspen-8',
        num_retries=30,
        interval=1,
        retrieve_execution=None,
    ):  # pylint: disable=too-many-arguments
        """
        Initialize the Backend object.

        Args:
            use_hardware (bool): If True, the code is run on one of the AWS Braket backends, by default on the Rigetti
                Aspen-8 chip (instead of using the AWS Braket SV1 Simulator)
            num_runs (int): Number of runs to collect statistics.  (default is 1000)
            verbose (bool): If True, statistics are printed, in addition to the measurement result being registered
                (at the end of the circuit).
            credentials (dict): mapping the AWS key credentials as the AWS_ACCESS_KEY_ID and AWS_SECRET_KEY.
            device (str): name of the device to use. Rigetti Aspen-8 by default. Valid names are "Aspen-8", "IonQ
                Device" and "SV1"
            num_retries (int): Number of times to retry to obtain results from AWS Braket. (default is 30)
            interval (float, int): Number of seconds between successive attempts to obtain results from AWS Braket.
                (default is 1)
            retrieve_execution (str): TaskArn to retrieve instead of re-running the circuit (e.g., if previous run
                timed out).  The TaskArns have the form:
                "arn:aws:braket:us-east-1:123456789012:quantum-task/5766032b-2b47-4bf9-cg00-f11851g4015b"
        """
        BasicEngine.__init__(self)
        self._reset()
        if use_hardware:
            self.device = device
        else:
            self.device = 'SV1'
        self._clear = False
        self._num_runs = num_runs
        self._verbose = verbose
        self._credentials = credentials
        self._s3_folder = s3_folder
        self._num_retries = num_retries
        self._interval = interval
        self._probabilities = dict()
        self._circuit = ""
        self._measured_ids = []
        self._allocated_qubits = set()
        self._retrieve_execution = retrieve_execution

        # Dictionary to translate the gates from ProjectQ to AWSBraket
        self._gationary = {
            XGate: 'x',
            YGate: 'y',
            ZGate: 'z',
            HGate: 'h',
            R: 'phaseshift',
            Rx: 'rx',
            Ry: 'ry',
            Rz: 'rz',
            SGate: 's',  # NB: Sdag is 'si'
            TGate: 't',  # NB: Tdag is 'ti'
            SwapGate: 'swap',
            SqrtXGate: 'v',
        }

        # Static head and tail to be added to the circuit
        # to build the "action".
        self._circuithead = '{"braketSchemaHeader": \
{"name": "braket.ir.jaqcd.program", "version": "1"}, \
"results": [], "basis_rotation_instructions": [], \
"instructions": ['

        self._circuittail = ']}'

    def is_available(self, cmd):  # pylint: disable=too-many-return-statements,too-many-branches
        """
        Return true if the command can be executed.

        Depending on the device chosen, the operations available differ.

        The operations avialable for the Aspen-8 Rigetti device are:
        - "cz" = Control Z, "xy" = Not available in ProjectQ, "ccnot" = Toffoli (ie. controlled CNOT), "cnot" =
          Control X, "cphaseshift" = Control R, "cphaseshift00" "cphaseshift01" "cphaseshift10" = Not available
          in ProjectQ,
          "cswap" = Control Swap, "h" = H, "i" = Identity, not in ProjectQ, "iswap" = Not available in ProjectQ,
          "phaseshift" = R, "pswap" = Not available in ProjectQ, "rx" = Rx, "ry" = Ry, "rz" = Rz, "s" = S, "si" =
          Sdag, "swap" = Swap, "t" = T, "ti" = Tdag, "x" = X, "y" = Y, "z" = Z

        The operations available for the IonQ Device are:
        - "x" = X, "y" = Y, "z" = Z, "rx" = Rx, "ry" = Ry, "rz" = Rz, "h", H, "cnot" = Control X, "s" = S, "si" =
          Sdag, "t" = T, "ti" = Tdag, "v" = SqrtX, "vi" = Not available in ProjectQ, "xx" "yy" "zz" = Not available in
          ProjectQ, "swap" = Swap, "i" = Identity, not in ProjectQ

        The operations available for the StateVector simulator (SV1) are the union of the ones for Rigetti Aspen-8 and
        IonQ Device plus some more:
        - "cy" = Control Y, "unitary" = Arbitrary unitary gate defined as a matrix equivalent to the MatrixGate in
          ProjectQ, "xy" = Not available in ProjectQ

        Args:
            cmd (Command): Command for which to check availability
        """

        gate = cmd.gate
        if gate in (Measure, Allocate, Deallocate, Barrier):
            return True

        if has_negative_control(cmd):
            return False

        if self.device == 'Aspen-8':
            if get_control_count(cmd) == 2:
                return isinstance(gate, XGate)
            if get_control_count(cmd) == 1:
                return isinstance(gate, (R, ZGate, XGate, SwapGate))
            if get_control_count(cmd) == 0:
                return (
                    isinstance(
                        gate,
                        (
                            R,
                            Rx,
                            Ry,
                            Rz,
                            XGate,
                            YGate,
                            ZGate,
                            HGate,
                            SGate,
                            TGate,
                            SwapGate,
                        ),
                    )
                    or gate in (Sdag, Tdag)
                )

        if self.device == 'IonQ Device':
            if get_control_count(cmd) == 1:
                return isinstance(gate, XGate)
            if get_control_count(cmd) == 0:
                return (
                    isinstance(
                        gate,
                        (
                            Rx,
                            Ry,
                            Rz,
                            XGate,
                            YGate,
                            ZGate,
                            HGate,
                            SGate,
                            TGate,
                            SqrtXGate,
                            SwapGate,
                        ),
                    )
                    or gate in (Sdag, Tdag)
                )

        if self.device == 'SV1':
            if get_control_count(cmd) == 2:
                return isinstance(gate, XGate)
            if get_control_count(cmd) == 1:
                return isinstance(gate, (R, ZGate, YGate, XGate, SwapGate))
            if get_control_count(cmd) == 0:
                # TODO: add MatrixGate to cover the unitary operation
                # TODO: Missing XY gate in ProjectQ
                return (
                    isinstance(
                        gate,
                        (
                            R,
                            Rx,
                            Ry,
                            Rz,
                            XGate,
                            YGate,
                            ZGate,
                            HGate,
                            SGate,
                            TGate,
                            SqrtXGate,
                            SwapGate,
                        ),
                    )
                    or gate in (Sdag, Tdag)
                )
        return False

    def _reset(self):
        """Reset all temporary variables (after flush gate)."""
        self._clear = True
        self._measured_ids = []

    def _store(self, cmd):  # pylint: disable=too-many-branches
        """
        Temporarily store the command cmd.

        Translates the command and stores it in a local variable (self._circuit) in JSON format.

        Args:
            cmd: Command to store
        """
        gate = cmd.gate

        # Do not clear the self._clear flag for those gates
        if gate in (Deallocate, Barrier):
            return

        num_controls = get_control_count(cmd)
        gate_type = (
            type(gate) if not isinstance(gate, DaggeredGate) else type(gate._gate)  # pylint: disable=protected-access
        )

        if self._clear:
            self._probabilities = dict()
            self._clear = False
            self._circuit = ""
            self._allocated_qubits = set()

        if gate == Allocate:
            self._allocated_qubits.add(cmd.qubits[0][0].id)
            return
        if gate == Measure:
            qb_id = cmd.qubits[0][0].id
            logical_id = None
            for tag in cmd.tags:
                if isinstance(tag, LogicalQubitIDTag):
                    logical_id = tag.logical_qubit_id
                    break
            self._measured_ids.append(logical_id if logical_id is not None else qb_id)
            return

        # All other supported gate types
        json_cmd = {}

        if num_controls > 1:
            json_cmd['controls'] = [qb.id for qb in cmd.control_qubits]
        elif num_controls == 1:
            json_cmd['control'] = cmd.control_qubits[0].id

        qubits = [qb.id for qureg in cmd.qubits for qb in qureg]
        if len(qubits) > 1:
            json_cmd['targets'] = qubits
        else:
            json_cmd['target'] = qubits[0]

        if isinstance(gate, (R, Rx, Ry, Rz)):
            json_cmd['angle'] = gate.angle

        if isinstance(gate, DaggeredGate):
            json_cmd['type'] = 'c' * num_controls + self._gationary[gate_type] + 'i'
        elif isinstance(gate, (XGate)) and num_controls > 0:
            json_cmd['type'] = 'c' * (num_controls - 1) + 'cnot'
        else:
            json_cmd['type'] = 'c' * num_controls + self._gationary[gate_type]

        self._circuit += json.dumps(json_cmd) + ", "

        # TODO: Add unitary for the SV1 simulator as MatrixGate

    def _logical_to_physical(self, qb_id):
        """
        Return the physical location of the qubit with the given logical id.

        Args:
            qb_id (int): ID of the logical qubit whose position should be
                returned.
        """
        if self.main_engine.mapper is not None:
            mapping = self.main_engine.mapper.current_mapping
            if qb_id not in mapping:
                raise RuntimeError(
                    (
                        "Unknown qubit id {} in current mapping. Please make sure eng.flush() was called and that the"
                        "qubit was eliminated during optimization."
                    ).format(qb_id)
                )
            return mapping[qb_id]
        return qb_id

    def get_probabilities(self, qureg):
        """
        Return the list of basis states with corresponding probabilities.  If input qureg is a subset of the register
        used for the experiment, then returns the projected probabilities over the other states.

        The measured bits are ordered according to the supplied quantum register, i.e., the left-most bit in the
        state-string corresponds to the first qubit in the supplied quantum register.

        Args:
            qureg (list<Qubit>): Quantum register determining the order of the
                qubits.

        Returns:
            probability_dict (dict): Dictionary mapping n-bit strings to
                probabilities.

        Raises:
            RuntimeError: If no data is available (i.e., if the circuit has not
                been executed). Or if a qubit was supplied which was not
                present in the circuit (might have gotten optimized away).

        Warning:
            Only call this function after the circuit has been executed!

            This is maintained in the same form of IBM and AQT for compatibility but in AWSBraket, a previously
            executed circuit will store the results in the S3 bucket and it can be retreived at any point in time
            thereafter.
            No circuit execution should be required at the time of retrieving the results and probabilities if the
            circuit has already been executed.
            In order to obtain the probabilities of a previous job you have to get the TaskArn and remember the qubits
            and ordering used in the original job.

        """
        if len(self._probabilities) == 0:
            raise RuntimeError("Please, run the circuit first!")

        probability_dict = dict()
        for state in self._probabilities:
            mapped_state = ['0'] * len(qureg)
            for i, qubit in enumerate(qureg):
                if self._logical_to_physical(qubit.id) >= len(state):  # pragma: no cover
                    raise IndexError('Physical ID {} > length of internal probabilities array'.format(qubit.id))
                mapped_state[i] = state[self._logical_to_physical(qubit.id)]
            probability = self._probabilities[state]
            mapped_state = "".join(mapped_state)
            if mapped_state not in probability_dict:
                probability_dict[mapped_state] = probability
            else:
                probability_dict[mapped_state] += probability
        return probability_dict

    def _run(self):
        """
        Run the circuit.

        Send the circuit via the AWS Boto3 SDK. Use the provided Access Key and Secret key or ask for them if not
        provided
        """
        # NB: the AWS Braket API does not require explicit measurement commands at the end of a circuit; after running
        # any circuit, all qubits are implicitly measured.  Also, AWS Braket currently does not support intermediate
        # measurements.

        # If the clear flag is set, nothing to do here...
        if self._clear:
            return

        # In Braket the results for the jobs are stored in S3.  You can recover the results from previous jobs using
        # the TaskArn (self._retrieve_execution).
        if self._retrieve_execution is not None:
            res = retrieve(
                credentials=self._credentials,
                task_arn=self._retrieve_execution,
                num_retries=self._num_retries,
                interval=self._interval,
                verbose=self._verbose,
            )
        else:
            # Return if no operations have been added.
            if not self._circuit:
                return

            n_qubit = len(self._allocated_qubits)
            info = {}
            info['circuit'] = self._circuithead + self._circuit.rstrip(', ') + self._circuittail
            info['nq'] = n_qubit
            info['shots'] = self._num_runs
            info['backend'] = {'name': self.device}
            res = send(
                info,
                device=self.device,
                credentials=self._credentials,
                s3_folder=self._s3_folder,
                num_retries=self._num_retries,
                interval=self._interval,
                verbose=self._verbose,
            )

        counts = res

        # Determine random outcome
        random_outcome = random.random()
        p_sum = 0.0
        measured = ""
        for state in counts:
            probability = counts[state]
            p_sum += probability
            star = ""
            if p_sum >= random_outcome and measured == "":
                measured = state
                star = "*"
            self._probabilities[state] = probability
            if self._verbose and probability > 0:
                print(state + " with p = " + str(probability) + star)

        # register measurement result
        for qubit_id in self._measured_ids:
            result = int(measured[self._logical_to_physical(qubit_id)])
            self.main_engine.set_measurement_result(WeakQubitRef(self.main_engine, qubit_id), result)
        self._reset()

    def receive(self, command_list):
        """
        Receives a command list and, for each command, stores it until completion.

        Args:
            command_list: List of commands to execute
        """
        for cmd in command_list:
            if not isinstance(cmd.gate, FlushGate):
                self._store(cmd)
            else:
                self._run()
                self._reset()
            if not self.is_last_engine:
                self.send([cmd])



# -*- coding: utf-8 -*-
#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""TestEngine and DummyEngine."""

from copy import deepcopy
from projectq.ops import FlushGate

from ._basics import BasicEngine


def _compare_cmds(cmd1, cmd2):
    """Compare two command objects"""
    cmd2 = deepcopy(cmd2)
    cmd2.engine = cmd1.engine
    return cmd1 == cmd2


class CompareEngine(BasicEngine):
    """
    CompareEngine is an engine which saves all commands. It is only intended for testing purposes. Two CompareEngine
    backends can be compared and return True if they contain the same commmands.
    """

    def __init__(self):
        super().__init__()
        self._l = [[]]

    def is_available(self, cmd):
        """All commands are accepted by this compiler engine"""
        return True

    def cache_cmd(self, cmd):
        """Cache a command"""
        # are there qubit ids that haven't been added to the list?
        all_qubit_id_list = [qubit.id for qureg in cmd.all_qubits for qubit in qureg]
        maxidx = int(0)
        for qubit_id in all_qubit_id_list:
            maxidx = max(maxidx, qubit_id)

        # if so, increase size of list to account for these qubits
        add = maxidx + 1 - len(self._l)
        if add > 0:
            for _ in range(add):
                self._l += [[]]

        # add gate command to each of the qubits involved
        for qubit_id in all_qubit_id_list:
            self._l[qubit_id] += [cmd]

    def receive(self, command_list):
        """
        Receives a command list and, for each command, stores it inside the cache before sending it to the next
        compiler engine.

        Args:
            command_list (list of Command objects): list of commands to receive.
        """
        for cmd in command_list:
            if not cmd.gate == FlushGate():
                self.cache_cmd(cmd)
        if not self.is_last_engine:
            self.send(command_list)

    def __eq__(self, other):
        if not isinstance(other, CompareEngine) or len(self._l) != len(other._l):
            return False
        for i in range(len(self._l)):
            if len(self._l[i]) != len(other._l[i]):
                return False
            for j in range(len(self._l[i])):
                if not _compare_cmds(self._l[i][j], other._l[i][j]):
                    return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        string = ""
        for qubit_id in range(len(self._l)):
            string += "Qubit {0} : ".format(qubit_id)
            for command in self._l[qubit_id]:
                string += str(command) + ", "
            string = string[:-2] + "\n"
        return string


class DummyEngine(BasicEngine):
    """
    DummyEngine used for testing.

    The DummyEngine forwards all commands directly to next engine.  If self.is_last_engine == True it just discards
    all gates.
    By setting save_commands == True all commands get saved as a list in self.received_commands. Elements are appended
    to this list so they are ordered according to when they are received.
    """

    def __init__(self, save_commands=False):
        """
        Initialize DummyEngine

        Args:
            save_commands (default = False): If True, commands are saved in
                                             self.received_commands.
        """
        super().__init__()
        self.save_commands = save_commands
        self.received_commands = []

    def is_available(self, cmd):
        """All commands are accepted by this compiler engine"""
        return True

    def receive(self, command_list):
        """
        Receives a command list and, for each command, stores it internally if requested before sending it to the next
        compiler engine.

        Args:
            command_list (list of Command objects): list of commands to receive.
        """
        if self.save_commands:
            self.received_commands.extend(command_list)
        if not self.is_last_engine:
            self.send(command_list)



# -*- coding: utf-8 -*-
# pylint: skip-file

from projectq import MainEngine
from projectq.backends import CircuitDrawer

import teleport

if __name__ == "__main__":
    # create a main compiler engine with a simulator backend:
    drawing_engine = CircuitDrawer()
    locations = {0: 1, 1: 2, 2: 0}
    drawing_engine.set_qubit_locations(locations)
    eng = MainEngine(drawing_engine)

    # we just want to draw the teleportation circuit
    def create_state(eng, qb):
        pass

    # run the teleport and then, let Bob try to uncompute his qubit:
    teleport.run_teleport(eng, create_state, verbose=False)

    # print latex code to draw the circuit:
    print(drawing_engine.get_latex())



# -*- coding: utf-8 -*-
#   Copyright 2021 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
# pylint: skip-file


from projectq.cengines import MainEngine
from projectq.meta import Control
from projectq.ops import All, X, Measure, CtrlAll


def run_circuit(eng, circuit_num):
    qubit = eng.allocate_qureg(2)
    ctrl_fail = eng.allocate_qureg(3)
    ctrl_success = eng.allocate_qureg(3)

    if circuit_num == 1:
        with Control(eng, ctrl_fail):
            X | qubit[0]
        All(X) | ctrl_success
        with Control(eng, ctrl_success):
            X | qubit[1]

    elif circuit_num == 2:
        All(X) | ctrl_fail
        with Control(eng, ctrl_fail, ctrl_state=CtrlAll.Zero):
            X | qubit[0]
        with Control(eng, ctrl_success, ctrl_state=CtrlAll.Zero):
            X | qubit[1]

    elif circuit_num == 3:
        All(X) | ctrl_fail
        with Control(eng, ctrl_fail, ctrl_state='101'):
            X | qubit[0]

        X | ctrl_success[0]
        X | ctrl_success[2]
        with Control(eng, ctrl_success, ctrl_state='101'):
            X | qubit[1]

    elif circuit_num == 4:
        All(X) | ctrl_fail
        with Control(eng, ctrl_fail, ctrl_state=5):
            X | qubit[0]

        X | ctrl_success[0]
        X | ctrl_success[2]
        with Control(eng, ctrl_success, ctrl_state=5):
            X | qubit[1]

    All(Measure) | qubit
    All(Measure) | ctrl_fail
    All(Measure) | ctrl_success
    eng.flush()
    return qubit, ctrl_fail, ctrl_success


if __name__ == '__main__':
    # Create a MainEngine with a unitary simulator backend
    eng = MainEngine()

    # Run out quantum circuit
    #   1 - Default behaviour of the control: all control qubits should be 1
    #   2 - Off-control: all control qubits should remain 0
    #   3 - Specific state given by a string
    #   4 - Specific state given by an integer

    qubit, ctrl_fail, ctrl_success = run_circuit(eng, 4)

    # Measured value of the failed qubit should be 0 in all cases
    print('The final value of the qubit with failed control is:')
    print(int(qubit[0]))
    print('with the state of control qubits are:')
    print([int(qubit) for qubit in ctrl_fail], '\n')

    # Measured value of the success qubit should be 1 in all cases
    print('The final value of the qubit with successful control is:')
    print(int(qubit[1]))
    print('with the state of control qubits are:')
    print([int(qubit) for qubit in ctrl_success], '\n')



# -*- coding: utf-8 -*-
#   Copyright 2021 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
Back-end to run quantum program on AWS Braket supported devices.

This backend requires the official AWS SDK for Python, Boto3.
The installation is very simple
> pip install boto3
"""

import getpass
import json
import re
import signal
import time

import boto3
import botocore


class AWSBraket:
    """
    Manage a session between ProjectQ and AWS Braket service.
    """

    def __init__(self):
        self.backends = dict()
        self.timeout = 5.0
        self._credentials = dict()
        self._s3_folder = []

    def authenticate(self, credentials=None):
        """
        Args:
            credentials (dict): mapping the AWS key credentials as the
                AWS_ACCESS_KEY_ID and AWS_SECRET_KEY.
        """
        if credentials is None:  # pragma: no cover
            credentials['AWS_ACCESS_KEY_ID'] = getpass.getpass(prompt="Enter AWS_ACCESS_KEY_ID: ")
            credentials['AWS_SECRET_KEY'] = getpass.getpass(prompt="Enter AWS_SECRET_KEY: ")

        self._credentials = credentials

    def get_s3_folder(self, s3_folder=None):
        """
        Args:
           s3_folder (list): contains the S3 bucket and directory to store the
           results.
        """
        if s3_folder is None:  # pragma: no cover
            s3_bucket = input("Enter the S3 Bucket configured in Braket: ")
            s3_directory = input("Enter the Directory created in the S3 Bucket: ")
            s3_folder = [s3_bucket, s3_directory]

        self._s3_folder = s3_folder

    def get_list_devices(self, verbose=False):
        """
        Get the list of available devices with their basic properties

        Args:
            verbose (bool): print the returned dictionnary if True

        Returns:
            (dict) backends dictionary by deviceName, containing the qubit size
                   'nq', the coupling map 'coupling_map' if applicable (IonQ
                   Device as an ion device is having full connectivity) and the
                   Schema Header version 'version', because it seems that no
                   device version is available by now
        """
        # TODO: refresh region_names if more regions get devices available
        self.backends = dict()
        region_names = ['us-west-1', 'us-east-1']
        for region in region_names:
            client = boto3.client(
                'braket',
                region_name=region,
                aws_access_key_id=self._credentials['AWS_ACCESS_KEY_ID'],
                aws_secret_access_key=self._credentials['AWS_SECRET_KEY'],
            )
            filters = []
            devicelist = client.search_devices(filters=filters)
            for result in devicelist['devices']:
                if result['deviceType'] not in ['QPU', 'SIMULATOR']:
                    continue
                if result['deviceType'] == 'QPU':
                    device_capabilities = json.loads(
                        client.get_device(deviceArn=result['deviceArn'])['deviceCapabilities']
                    )
                    self.backends[result['deviceName']] = {
                        'nq': device_capabilities['paradigm']['qubitCount'],
                        'coupling_map': device_capabilities['paradigm']['connectivity']['connectivityGraph'],
                        'version': device_capabilities['braketSchemaHeader']['version'],
                        'location': region,  # deviceCapabilities['service']['deviceLocation'],
                        'deviceArn': result['deviceArn'],
                        'deviceParameters': device_capabilities['deviceParameters']['properties']['braketSchemaHeader'][
                            'const'
                        ],
                        'deviceModelParameters': device_capabilities['deviceParameters']['definitions'][
                            'GateModelParameters'
                        ]['properties']['braketSchemaHeader']['const'],
                    }
                # Unfortunatelly the Capabilities schemas are not homogeneus
                # for real devices and simulators
                elif result['deviceType'] == 'SIMULATOR':
                    device_capabilities = json.loads(
                        client.get_device(deviceArn=result['deviceArn'])['deviceCapabilities']
                    )
                    self.backends[result['deviceName']] = {
                        'nq': device_capabilities['paradigm']['qubitCount'],
                        'coupling_map': {},
                        'version': device_capabilities['braketSchemaHeader']['version'],
                        'location': 'us-east-1',
                        'deviceArn': result['deviceArn'],
                        'deviceParameters': device_capabilities['deviceParameters']['properties']['braketSchemaHeader'][
                            'const'
                        ],
                        'deviceModelParameters': device_capabilities['deviceParameters']['definitions'][
                            'GateModelParameters'
                        ]['properties']['braketSchemaHeader']['const'],
                    }

        if verbose:
            print('- List of AWSBraket devices available:')
            print(list(self.backends))

        return self.backends

    def is_online(self, device):
        """
        Check if the device is in the list of available backends.

        Args:
            device (str): name of the device to check

        Returns:
            (bool) True if device is available, False otherwise
        """
        # TODO: Add info for the device if it is actually ONLINE
        return device in self.backends

    def can_run_experiment(self, info, device):
        """
        Check if the device is big enough to run the code.

        Args:
            info (dict): dictionary sent by the backend containing the code to
                run
            device (str): name of the device to use

        Returns:
            (tuple): (bool) True if device is big enough, False otherwise (int)
                     maximum number of qubit available on the device (int)
                     number of qubit needed for the circuit

        """
        nb_qubit_max = self.backends[device]['nq']
        nb_qubit_needed = info['nq']
        return nb_qubit_needed <= nb_qubit_max, nb_qubit_max, nb_qubit_needed

    def run(self, info, device):
        """
        Run the quantum code to the AWS Braket selected device.

        Args:
            info (dict): dictionary sent by the backend containing the code to
                run
            device (str): name of the device to use

        Returns:
            task_arn (str): The Arn of the task


        """
        argument = {
            'circ': info['circuit'],
            's3_folder': self._s3_folder,
            'shots': info['shots'],
        }

        region_name = self.backends[device]['location']
        device_parameters = {
            'braketSchemaHeader': self.backends[device]['deviceParameters'],
            'paradigmParameters': {
                'braketSchemaHeader': self.backends[device]['deviceModelParameters'],
                'qubitCount': info['nq'],
                'disableQubitRewiring': False,
            },
        }
        device_parameters = json.dumps(device_parameters)

        client_braket = boto3.client(
            'braket',
            region_name=region_name,
            aws_access_key_id=self._credentials['AWS_ACCESS_KEY_ID'],
            aws_secret_access_key=self._credentials['AWS_SECRET_KEY'],
        )

        response = client_braket.create_quantum_task(
            action=argument['circ'],
            deviceArn=self.backends[device]['deviceArn'],
            deviceParameters=device_parameters,
            outputS3Bucket=argument['s3_folder'][0],
            outputS3KeyPrefix=argument['s3_folder'][1],
            shots=argument['shots'],
        )

        return response['quantumTaskArn']

    def get_result(self, execution_id, num_retries=30, interval=1, verbose=False):  # pylint: disable=too-many-locals
        """
        Get the result of an execution
        """
        if verbose:
            print("Waiting for results. [Job Arn: {}]".format(execution_id))

        original_sigint_handler = signal.getsignal(signal.SIGINT)

        def _handle_sigint_during_get_result(*_):  # pragma: no cover
            raise Exception("Interrupted. The Arn of your submitted job is {}.".format(execution_id))

        def _calculate_measurement_probs(measurements):
            """
            Calculate the measurement probabilities based on the
            list of measurements for a job sent to a SV1 Braket simulator

            Args:
                measurements (list): list of measurements

            Returns:
                measurementsProbabilities (dict): The measurements
                with their probabilities
            """
            total_mes = len(measurements)
            unique_mes = [list(x) for x in set(tuple(x) for x in measurements)]
            total_unique_mes = len(unique_mes)
            len_qubits = len(unique_mes[0])
            measurements_probabilities = {}
            for i in range(total_unique_mes):
                strqubits = ''
                for qubit_idx in range(len_qubits):
                    strqubits += str(unique_mes[i][qubit_idx])
                prob = measurements.count(unique_mes[i]) / total_mes
                measurements_probabilities[strqubits] = prob

            return measurements_probabilities

        # The region_name is obtained from the task_arn itself
        region_name = re.split(':', execution_id)[3]
        client_braket = boto3.client(
            'braket',
            region_name=region_name,
            aws_access_key_id=self._credentials['AWS_ACCESS_KEY_ID'],
            aws_secret_access_key=self._credentials['AWS_SECRET_KEY'],
        )

        try:
            signal.signal(signal.SIGINT, _handle_sigint_during_get_result)

            for _ in range(num_retries):
                quantum_task = client_braket.get_quantum_task(quantumTaskArn=execution_id)
                status = quantum_task['status']
                bucket = quantum_task['outputS3Bucket']
                directory = quantum_task['outputS3Directory']
                resultsojectname = directory + '/results.json'
                if status == 'COMPLETED':
                    # Get the device type to obtian the correct measurement
                    # structure
                    devicetype_used = client_braket.get_device(deviceArn=quantum_task['deviceArn'])['deviceType']
                    # Get the results from S3
                    client_s3 = boto3.client(
                        's3',
                        aws_access_key_id=self._credentials['AWS_ACCESS_KEY_ID'],
                        aws_secret_access_key=self._credentials['AWS_SECRET_KEY'],
                    )
                    s3result = client_s3.get_object(Bucket=bucket, Key=resultsojectname)
                    if verbose:
                        print("Results obtained. [Status: {}]".format(status))
                    result_content = json.loads(s3result['Body'].read())

                    if devicetype_used == 'QPU':
                        return result_content['measurementProbabilities']
                    if devicetype_used == 'SIMULATOR':
                        return _calculate_measurement_probs(result_content['measurements'])
                if status == 'FAILED':
                    raise Exception(
                        "Error while running the code: {}. "
                        "The failure reason was: {}.".format(status, quantum_task['failureReason'])
                    )
                if status == 'CANCELLING':
                    raise Exception("The job received a CANCEL operation: {}.".format(status))
                time.sleep(interval)
                # NOTE: Be aware that AWS is billing if a lot of API calls are
                # executed, therefore the num_repetitions is set to a small
                # number by default.
                # For QPU devices the job is always queued and there are some
                # working hours available.
                # In addition the results and state is writen in the
                # results.json file in the S3 Bucket and does not depend on the
                # status of the device

        finally:
            if original_sigint_handler is not None:
                signal.signal(signal.SIGINT, original_sigint_handler)

        raise Exception(
            "Timeout. "
            "The Arn of your submitted job is {} and the status "
            "of the job is {}.".format(execution_id, status)
        )


class DeviceTooSmall(Exception):
    """Exception raised if the device is too small to run the circuit"""


class DeviceOfflineError(Exception):
    """Exception raised if a selected device is currently offline"""


def show_devices(credentials=None, verbose=False):
    """
    Access the list of available devices and their properties (ex: for setup
    configuration)

    Args:
        credentials (dict): Dictionary storing the AWS credentials with
            keys AWS_ACCESS_KEY_ID and AWS_SECRET_KEY.
        verbose (bool): If True, additional information is printed

    Returns:
        (list) list of available devices and their properties
    """
    awsbraket_session = AWSBraket()
    awsbraket_session.authenticate(credentials=credentials)
    return awsbraket_session.get_list_devices(verbose=verbose)


# TODO: Create a Show Online properties per device


def retrieve(credentials, task_arn, num_retries=30, interval=1, verbose=False):
    """
    Retrieves a job/task by its Arn.

    Args:
        credentials (dict): Dictionary storing the AWS credentials with
            keys AWS_ACCESS_KEY_ID and AWS_SECRET_KEY.
        task_arn (str): The Arn of the task to retreive

    Returns:
        (dict) measurement probabilities from the result
        stored in the S3 folder
    """
    try:
        awsbraket_session = AWSBraket()
        if verbose:
            print("- Authenticating...")
            if credentials is not None:
                print("AWS credentials: " + credentials['AWS_ACCESS_KEY_ID'] + ", " + credentials['AWS_SECRET_KEY'])
        awsbraket_session.authenticate(credentials=credentials)
        res = awsbraket_session.get_result(task_arn, num_retries=num_retries, interval=interval, verbose=verbose)
        return res
    except botocore.exceptions.ClientError as error:
        error_code = error.response['Error']['Code']
        if error_code == 'ResourceNotFoundException':
            print("- Unable to locate the job with Arn ", task_arn)
        print(error, error_code)
        raise


def send(  # pylint: disable=too-many-branches,too-many-arguments,too-many-locals
    info, device, credentials, s3_folder, num_retries=30, interval=1, verbose=False
):
    """
    Sends cicruit through the Boto3 SDK and runs the quantum circuit.

    Args:
        info(dict): Contains representation of the circuit to run.
        device (str): name of the AWS Braket device.
        credentials (dict): Dictionary storing the AWS credentials with keys
            AWS_ACCESS_KEY_ID and AWS_SECRET_KEY.
        s3_folder (list): Contains the S3 bucket and directory to store the
            results.
        verbose (bool): If True, additional information is printed, such as
            measurement statistics. Otherwise, the backend simply registers one
            measurement result (same behavior as the projectq Simulator).

    Returns:
        (list) samples from the AWS Braket device

    """
    try:
        awsbraket_session = AWSBraket()
        if verbose:
            print("- Authenticating...")
            if credentials is not None:
                print("AWS credentials: " + credentials['AWS_ACCESS_KEY_ID'] + ", " + credentials['AWS_SECRET_KEY'])
        awsbraket_session.authenticate(credentials=credentials)
        awsbraket_session.get_s3_folder(s3_folder=s3_folder)

        # check if the device is online/is available
        awsbraket_session.get_list_devices(verbose)
        online = awsbraket_session.is_online(device)
        if online:
            print("The job will be queued in any case, plase take this into account")
        else:
            print("The device is not available. Use the simulator instead or try another device.")
            raise DeviceOfflineError("Device is not available.")

        # check if the device has enough qubit to run the code
        runnable, qmax, qneeded = awsbraket_session.can_run_experiment(info, device)
        if not runnable:
            print(
                (
                    "The device is too small ({} qubits available) for the code "
                    + "requested({} qubits needed) Try to look for another "
                    + "device with more qubits"
                ).format(qmax, qneeded)
            )
            raise DeviceTooSmall("Device is too small.")
        if verbose:
            print("- Running code: {}".format(info))
        task_arn = awsbraket_session.run(info, device)
        print("Your task Arn is: {}. Make note of that for future reference".format(task_arn))

        if verbose:
            print("- Waiting for results...")
        res = awsbraket_session.get_result(task_arn, num_retries=num_retries, interval=interval, verbose=verbose)
        if verbose:
            print("- Done.")
        return res

    except botocore.exceptions.ClientError as error:
        error_code = error.response['Error']['Code']
        if error_code == 'AccessDeniedException':
            print("- There was an error: the access to Braket was denied")
        if error_code == 'DeviceOfflineException':
            print("- There was an error: the device is offline")
        if error_code == 'InternalServiceException':
            print("- There was an interal Bracket service error")
        if error_code == 'ServiceQuotaExceededException':
            print("- There was an error: the quota on Braket was exceed")
        if error_code == 'ValidationException':
            print("- There was a Validation error")
        print(error, error_code)
        raise



# -*- coding: utf-8 -*-
#   Copyright 2018 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
Registers decomposition for StatePreparation.
"""

import cmath
import math

from projectq.cengines import DecompositionRule
from projectq.meta import Control, Dagger
from projectq.ops import (
    StatePreparation,
    UniformlyControlledRy,
    UniformlyControlledRz,
    Ph,
)


def _decompose_state_preparation(cmd):  # pylint: disable=too-many-locals
    """
    Implements state preparation based on arXiv:quant-ph/0407010v1.
    """
    eng = cmd.engine
    if len(cmd.qubits) != 1:
        raise ValueError('StatePreparation does not support multiple quantum registers!')
    num_qubits = len(cmd.qubits[0])
    qureg = cmd.qubits[0]
    final_state = cmd.gate.final_state
    if len(final_state) != 2 ** num_qubits:
        raise ValueError("Length of final_state is invalid.")
    norm = 0.0
    for amplitude in final_state:
        norm += abs(amplitude) ** 2
    if norm < 1 - 1e-10 or norm > 1 + 1e-10:
        raise ValueError("final_state is not normalized.")
    with Control(eng, cmd.control_qubits):
        # As in the paper reference, we implement the inverse:
        with Dagger(eng):
            # Cancel all the relative phases
            phase_of_blocks = []
            for amplitude in final_state:
                phase_of_blocks.append(cmath.phase(amplitude))
            for qubit_idx, qubit in enumerate(qureg):
                angles = []
                phase_of_next_blocks = []
                for block in range(2 ** (len(qureg) - qubit_idx - 1)):
                    phase0 = phase_of_blocks[2 * block]
                    phase1 = phase_of_blocks[2 * block + 1]
                    angles.append(phase0 - phase1)
                    phase_of_next_blocks.append((phase0 + phase1) / 2.0)
                UniformlyControlledRz(angles) | (
                    qureg[(qubit_idx + 1) :],  # noqa: E203
                    qubit,
                )
                phase_of_blocks = phase_of_next_blocks
            # Cancel global phase
            Ph(-phase_of_blocks[0]) | qureg[-1]
            # Remove amplitudes from states which contain a bit value 1:
            abs_of_blocks = []
            for amplitude in final_state:
                abs_of_blocks.append(abs(amplitude))
            for qubit_idx, qubit in enumerate(qureg):
                angles = []
                abs_of_next_blocks = []
                for block in range(2 ** (len(qureg) - qubit_idx - 1)):
                    a0 = abs_of_blocks[2 * block]  # pylint: disable=invalid-name
                    a1 = abs_of_blocks[2 * block + 1]  # pylint: disable=invalid-name
                    if a0 == 0 and a1 == 0:
                        angles.append(0)
                    else:
                        angles.append(-2.0 * math.acos(a0 / math.sqrt(a0 ** 2 + a1 ** 2)))
                    abs_of_next_blocks.append(math.sqrt(a0 ** 2 + a1 ** 2))
                UniformlyControlledRy(angles) | (
                    qureg[(qubit_idx + 1) :],  # noqa: E203
                    qubit,
                )
                abs_of_blocks = abs_of_next_blocks


#: Decomposition rules
all_defined_decomposition_rules = [DecompositionRule(StatePreparation, _decompose_state_preparation)]



# -*- coding: utf-8 -*-
#   Copyright 2018 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
Defines the parent class from which all mappers should be derived.

There is only one engine currently allowed to be derived from BasicMapperEngine. This allows the simulator to
automatically translate logical qubit ids to mapped ids.
"""
from copy import deepcopy

from projectq.meta import drop_engine_after, insert_engine, LogicalQubitIDTag
from projectq.ops import MeasureGate

from ._basics import BasicEngine
from ._cmdmodifier import CommandModifier


class BasicMapperEngine(BasicEngine):
    """
    Parent class for all Mappers.

    Attributes:
        self.current_mapping (dict): Keys are the logical qubit ids and values are the mapped qubit ids.

    """

    def __init__(self):
        super().__init__()
        self._current_mapping = None

    @property
    def current_mapping(self):
        """
        Access the current mapping
        """
        return deepcopy(self._current_mapping)

    @current_mapping.setter
    def current_mapping(self, current_mapping):
        """
        Set the current mapping
        """
        self._current_mapping = current_mapping

    def _send_cmd_with_mapped_ids(self, cmd):
        """
        Send this Command using the mapped qubit ids of self.current_mapping.

        If it is a Measurement gate, then it adds a LogicalQubitID tag.

        Args:
            cmd: Command object with logical qubit ids.
        """
        new_cmd = deepcopy(cmd)
        qubits = new_cmd.qubits
        for qureg in qubits:
            for qubit in qureg:
                if qubit.id != -1:
                    qubit.id = self.current_mapping[qubit.id]
        control_qubits = new_cmd.control_qubits
        for qubit in control_qubits:
            qubit.id = self.current_mapping[qubit.id]
        if isinstance(new_cmd.gate, MeasureGate):
            # Add LogicalQubitIDTag to MeasureGate
            def add_logical_id(command, old_tags=deepcopy(cmd.tags)):
                command.tags = old_tags + [LogicalQubitIDTag(cmd.qubits[0][0].id)]
                return command

            tagger_eng = CommandModifier(add_logical_id)
            insert_engine(self, tagger_eng)
            self.send([new_cmd])
            drop_engine_after(self)
        else:
            self.send([new_cmd])

    def receive(self, command_list):
        """Forward all commands to the next engine."""
        for cmd in command_list:
            self._send_cmd_with_mapped_ids(cmd)



# -*- coding: utf-8 -*-
#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Module containing some helper classes for generating the documentation"""

import inspect
import sys
import os


class PackageDescription:  # pylint: disable=too-many-instance-attributes,too-few-public-methods
    """Class representing a package description"""

    package_list = []

    def __init__(  # pylint: disable=too-many-arguments
        self,
        pkg_name,
        desc='',
        module_special_members='__init__',
        submodule_special_members='',
        submodules_desc='',
        helper_submodules=None,
    ):
        """
        Args:
            name (str): Name of ProjectQ module
            desc (str): (optional) Description of module
            module_special_members (str): (optional) Special members to include
                in the documentation of the module
            submodule_special_members (str): (optional) Special members to
                include in the documentation of submodules
            submodules_desc (str): (optional) Description to print out before
                the list of submodules
            helper_submodules (list): (optional) List of tuples for helper
                sub-modules to include in the documentation.
                Tuples are (section_title, submodukle_name,
                            automodule_properties)
        """

        self.name = pkg_name
        self.desc = desc
        if pkg_name not in PackageDescription.package_list:
            PackageDescription.package_list.append(pkg_name)

        self.module = sys.modules['projectq.{}'.format(self.name)]
        self.module_special_members = module_special_members

        self.submodule_special_members = submodule_special_members
        self.submodules_desc = submodules_desc

        self.helper_submodules = helper_submodules

        module_root = os.path.dirname(self.module.__file__)
        sub = [
            (name, obj)
            for name, obj in inspect.getmembers(
                self.module,
                lambda obj: inspect.ismodule(obj) and hasattr(obj, '__file__') and module_root in obj.__file__,
            )
            if pkg_name[0] != '_'
        ]

        self.subpackages = []
        self.submodules = []
        for name, obj in sub:
            if '{}.{}'.format(self.name, name) in PackageDescription.package_list:
                self.subpackages.append((name, obj))
            else:
                self.submodules.append((name, obj))

        self.subpackages.sort(key=lambda x: x[0].lower())
        self.submodules.sort(key=lambda x: x[0].lower())

        self.members = [
            (name, obj)
            for name, obj in inspect.getmembers(
                self.module,
                lambda obj: (
                    inspect.isclass(obj)
                    or inspect.isfunction(obj)
                    or isinstance(obj, (int, float, tuple, list, dict, set, frozenset, str))
                ),
            )
            if name[0] != '_'
        ]
        self.members.sort(key=lambda x: x[0].lower())

    def get_ReST(self):  # pylint: disable=invalid-name,too-many-branches,too-many-statements
        """
        Conversion to ReST formatted string.
        """
        new_lines = []
        new_lines.append(self.name)
        new_lines.append('=' * len(self.name))
        new_lines.append('')

        if self.desc:
            new_lines.append(self.desc.strip())
            new_lines.append('')

        submodule_has_index = False

        if self.subpackages:
            new_lines.append('Subpackages')
            new_lines.append('-' * len(new_lines[-1]))
            new_lines.append('')
            new_lines.append('.. toctree::')
            new_lines.append('    :maxdepth: 1')
            new_lines.append('')
            for name, _ in self.subpackages:
                new_lines.append('    projectq.{}.{}'.format(self.name, name))
            new_lines.append('')
        else:
            submodule_has_index = True
            new_lines.append('.. autosummary::')
            new_lines.append('')
            if self.submodules:
                for name, _ in self.submodules:
                    new_lines.append('\tprojectq.{}.{}'.format(self.name, name))
                new_lines.append('')
            if self.members:
                for name, _ in self.members:
                    new_lines.append('\tprojectq.{}.{}'.format(self.name, name))
                new_lines.append('')

        if self.submodules:
            new_lines.append('Submodules')
            new_lines.append('-' * len(new_lines[-1]))
            new_lines.append('')
            if self.submodules_desc:
                new_lines.append(self.submodules_desc.strip())
                new_lines.append('')

            if not submodule_has_index:
                new_lines.append('.. autosummary::')
                new_lines.append('')
                for name, _ in self.submodules:
                    new_lines.append('    projectq.{}.{}'.format(self.name, name))
                new_lines.append('')

            for name, _ in self.submodules:
                new_lines.append(name)
                new_lines.append('^' * len(new_lines[-1]))
                new_lines.append('')
                new_lines.append('.. automodule:: projectq.{}.{}'.format(self.name, name))
                new_lines.append('    :members:')
                if self.submodule_special_members:
                    new_lines.append('    :special-members: {}'.format(self.submodule_special_members))
                new_lines.append('    :undoc-members:')
                new_lines.append('')

        new_lines.append('Module contents')
        new_lines.append('-' * len(new_lines[-1]))
        new_lines.append('')
        new_lines.append('.. automodule:: projectq.{}'.format(self.name))
        new_lines.append('    :members:')
        new_lines.append('    :undoc-members:')
        new_lines.append('    :special-members: {}'.format(self.module_special_members))
        new_lines.append('    :imported-members:')
        new_lines.append('')

        if self.helper_submodules:
            new_lines.append('Helper sub-modules')
            new_lines.append('-' * len(new_lines[-1]))
            new_lines.append('')
            for title, name, params in self.helper_submodules:
                new_lines.append(title)
                new_lines.append('^' * len(title))
                new_lines.append('')
                new_lines.append('.. automodule:: projectq.{}.{}'.format(self.name, name))
                for param in params:
                    new_lines.append('    {}'.format(param))
                new_lines.append('')

        return new_lines[:-1]



# -*- coding: utf-8 -*-
#   Copyright 2021 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Mapper that has a max number of allocatable qubits."""
from projectq.cengines import BasicMapperEngine
from projectq.meta import LogicalQubitIDTag
from projectq.ops import AllocateQubitGate, Command, DeallocateQubitGate, FlushGate
from projectq.types import WeakQubitRef


class BoundedQubitMapper(BasicMapperEngine):
    """Maps logical qubits to a fixed number of hardware qubits"""

    def __init__(self, max_qubits):
        super().__init__()
        self._qubit_idx = 0
        self.max_qubits = max_qubits

    def _reset(self):
        # Reset the mapping index.
        self._qubit_idx = 0

    def _process_cmd(self, cmd):
        current_mapping = self.current_mapping
        if current_mapping is None:
            current_mapping = dict()

        if isinstance(cmd.gate, AllocateQubitGate):
            qubit_id = cmd.qubits[0][0].id
            if qubit_id in current_mapping:
                raise RuntimeError("Qubit with id {} has already been allocated!".format(qubit_id))

            if self._qubit_idx >= self.max_qubits:
                raise RuntimeError("Cannot allocate more than {} qubits!".format(self.max_qubits))

            new_id = self._qubit_idx
            self._qubit_idx += 1
            current_mapping[qubit_id] = new_id
            qb = WeakQubitRef(engine=self, idx=new_id)
            new_cmd = Command(
                engine=self,
                gate=AllocateQubitGate(),
                qubits=([qb],),
                tags=[LogicalQubitIDTag(qubit_id)],
            )
            self.current_mapping = current_mapping
            self.send([new_cmd])
        elif isinstance(cmd.gate, DeallocateQubitGate):
            qubit_id = cmd.qubits[0][0].id
            if qubit_id not in current_mapping:
                raise RuntimeError("Cannot deallocate a qubit that is not already allocated!")
            qb = WeakQubitRef(engine=self, idx=current_mapping[qubit_id])
            new_cmd = Command(
                engine=self,
                gate=DeallocateQubitGate(),
                qubits=([qb],),
                tags=[LogicalQubitIDTag(qubit_id)],
            )
            current_mapping.pop(qubit_id)
            self.current_mapping = current_mapping
            self.send([new_cmd])
        else:
            self._send_cmd_with_mapped_ids(cmd)

    def receive(self, command_list):
        for cmd in command_list:
            if isinstance(cmd.gate, FlushGate):
                self._reset()
                self.send([cmd])
            else:
                self._process_cmd(cmd)


__all__ = ['BoundedQubitMapper']



# -*- coding: utf-8 -*-
#   Copyright 2018 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for projectq.cengines._2dmapper.py."""

from copy import deepcopy
import itertools
import random

import pytest

import projectq
from projectq.cengines import DummyEngine, LocalOptimizer
from projectq.meta import LogicalQubitIDTag
from projectq.ops import Allocate, BasicGate, Command, Deallocate, FlushGate, X
from projectq.types import WeakQubitRef

from projectq.cengines import _twodmapper as two_d


def test_is_available():
    mapper = two_d.GridMapper(num_rows=2, num_columns=2)
    qb0 = WeakQubitRef(engine=None, idx=0)
    qb1 = WeakQubitRef(engine=None, idx=1)
    qb2 = WeakQubitRef(engine=None, idx=2)
    cmd0 = Command(None, BasicGate(), qubits=([qb0],))
    assert mapper.is_available(cmd0)
    cmd1 = Command(None, BasicGate(), qubits=([qb0],), controls=[qb1])
    assert mapper.is_available(cmd1)
    cmd2 = Command(None, BasicGate(), qubits=([qb0], [qb1, qb2]))
    assert not mapper.is_available(cmd2)
    cmd3 = Command(None, BasicGate(), qubits=([qb0], [qb1]), controls=[qb2])
    assert not mapper.is_available(cmd3)


def test_wrong_init_mapped_ids_to_backend_ids():
    with pytest.raises(RuntimeError):
        test = {0: 1, 1: 0, 2: 2, 3: 3, 4: 4}
        two_d.GridMapper(num_rows=2, num_columns=3, mapped_ids_to_backend_ids=test)
    with pytest.raises(RuntimeError):
        test = {0: 1, 1: 0, 2: 2, 3: 3, 4: 4, 5: 2}
        two_d.GridMapper(num_rows=2, num_columns=3, mapped_ids_to_backend_ids=test)


def test_resetting_mapping_to_none():
    mapper = two_d.GridMapper(num_rows=2, num_columns=3)
    mapper.current_mapping = {0: 1}
    assert mapper._current_row_major_mapping == {0: 1}
    mapper.current_mapping = None
    assert mapper._current_row_major_mapping is None


@pytest.mark.parametrize("different_backend_ids", [False, True])
def test_return_new_mapping(different_backend_ids):
    if different_backend_ids:
        map_to_backend_ids = {
            0: 21,
            1: 32,
            2: 1,
            3: 4,
            4: 5,
            5: 6,
            6: 10,
            7: 7,
            8: 0,
            9: 56,
            10: 55,
            11: 9,
        }
    else:
        map_to_backend_ids = None
    mapper = two_d.GridMapper(num_rows=4, num_columns=3, mapped_ids_to_backend_ids=map_to_backend_ids)
    eng = projectq.MainEngine(DummyEngine(), [mapper])
    linear_chain_ids = [33, 22, 11, 2, 3, 0, 6, 7, 9, 12, 4, 88]
    mapper._stored_commands = []
    for i in range(12):
        qb = WeakQubitRef(engine=None, idx=linear_chain_ids[i])
        cmd = Command(None, Allocate, ([qb],))
        mapper._stored_commands.append(cmd)
    for i in range(11):
        qb0 = WeakQubitRef(engine=None, idx=linear_chain_ids[i])
        qb1 = WeakQubitRef(engine=None, idx=linear_chain_ids[i + 1])
        cmd = Command(None, X, qubits=([qb0],), controls=[qb1])
        mapper._stored_commands.append(cmd)
    new_mapping = mapper._return_new_mapping()
    possible_solution_1 = {
        33: 0,
        22: 1,
        11: 2,
        2: 5,
        3: 4,
        0: 3,
        6: 6,
        7: 7,
        9: 8,
        12: 11,
        4: 10,
        88: 9,
    }
    possible_solution_2 = {
        88: 0,
        4: 1,
        12: 2,
        9: 5,
        7: 4,
        6: 3,
        0: 6,
        3: 7,
        2: 8,
        11: 11,
        22: 10,
        33: 9,
    }
    assert new_mapping == possible_solution_1 or new_mapping == possible_solution_2
    eng.flush()
    if different_backend_ids:
        transformed_sol1 = dict()
        for logical_id, mapped_id in possible_solution_1.items():
            transformed_sol1[logical_id] = map_to_backend_ids[mapped_id]
        transformed_sol2 = dict()
        for logical_id, mapped_id in possible_solution_2.items():
            transformed_sol2[logical_id] = map_to_backend_ids[mapped_id]
        assert mapper.current_mapping == transformed_sol1 or mapper.current_mapping == transformed_sol2
    else:
        assert mapper.current_mapping == possible_solution_1 or mapper.current_mapping == possible_solution_2


@pytest.mark.parametrize(
    "num_rows, num_columns, seed, none_old, none_new",
    [
        (2, 2, 0, 0, 0),
        (3, 4, 1, 0, 0),
        (4, 3, 2, 0, 0),
        (5, 5, 3, 0, 0),
        (5, 3, 4, 3, 0),
        (4, 4, 5, 0, 3),
        (6, 6, 7, 2, 3),
    ],
)
def test_return_swaps_random(num_rows, num_columns, seed, none_old, none_new):
    random.seed(seed)
    num_qubits = num_rows * num_columns
    old_chain = random.sample(range(num_qubits), num_qubits)
    new_chain = random.sample(range(num_qubits), num_qubits)
    old_mapping = dict()
    new_mapping = dict()
    for i in range(num_qubits):
        old_mapping[old_chain[i]] = i
        new_mapping[new_chain[i]] = i
    # Remove certain elements from mappings:
    old_none_ids = set(random.sample(range(num_qubits), none_old))
    if none_old != 0:
        for logical_id in old_none_ids:
            old_mapping.pop(logical_id)
    new_none_ids = set(random.sample(range(num_qubits), none_new))
    if none_new != 0:
        for logical_id in new_none_ids:
            new_mapping.pop(logical_id)

    mapper = two_d.GridMapper(num_rows=num_rows, num_columns=num_columns)
    swaps = mapper.return_swaps(old_mapping, new_mapping)
    # Check that Swaps are allowed
    all_allowed_swaps = set()
    for row in range(num_rows):
        for column in range(num_columns - 1):
            qb_id = row * num_columns + column
            all_allowed_swaps.add((qb_id, qb_id + 1))
    for row in range(num_rows - 1):
        for column in range(num_columns):
            qb_id = row * num_columns + column
            all_allowed_swaps.add((qb_id, qb_id + num_columns))

    for swap in swaps:
        assert swap in all_allowed_swaps
    test_chain = deepcopy(old_chain)
    for pos0, pos1 in swaps:
        tmp = test_chain[pos0]
        test_chain[pos0] = test_chain[pos1]
        test_chain[pos1] = tmp
    assert len(test_chain) == len(new_chain)
    for i in range(len(new_chain)):
        if new_chain[i] in old_mapping and new_chain[i] in new_mapping:
            assert test_chain[i] == new_chain[i]


@pytest.mark.parametrize("different_backend_ids", [False, True])
def test_send_possible_commands(different_backend_ids):
    if different_backend_ids:
        map_to_backend_ids = {0: 21, 1: 32, 2: 1, 3: 4, 4: 5, 5: 6, 6: 10, 7: 7}
    else:
        map_to_backend_ids = None
    mapper = two_d.GridMapper(num_rows=2, num_columns=4, mapped_ids_to_backend_ids=map_to_backend_ids)
    backend = DummyEngine(save_commands=True)
    backend.is_last_engine = True
    mapper.next_engine = backend
    # mapping is identical except 5 <-> 0
    if different_backend_ids:
        mapper.current_mapping = {0: 6, 1: 32, 2: 1, 3: 4, 4: 5, 5: 21, 6: 10, 7: 7}
    else:
        mapper.current_mapping = {5: 0, 1: 1, 2: 2, 3: 3, 4: 4, 0: 5, 6: 6, 7: 7}
    neighbours = [
        (5, 1),
        (1, 2),
        (2, 3),
        (4, 0),
        (0, 6),
        (6, 7),
        (5, 4),
        (1, 0),
        (2, 6),
        (3, 7),
    ]
    for qb0_id, qb1_id in neighbours:
        qb0 = WeakQubitRef(engine=None, idx=qb0_id)
        qb1 = WeakQubitRef(engine=None, idx=qb1_id)
        cmd1 = Command(None, X, qubits=([qb0],), controls=[qb1])
        cmd2 = Command(None, X, qubits=([qb1],), controls=[qb0])
        mapper._stored_commands = [cmd1, cmd2]
        mapper._send_possible_commands()
        assert len(mapper._stored_commands) == 0
    for qb0_id, qb1_id in itertools.permutations(range(8), 2):
        if (qb0_id, qb1_id) not in neighbours and (qb1_id, qb0_id) not in neighbours:
            qb0 = WeakQubitRef(engine=None, idx=qb0_id)
            qb1 = WeakQubitRef(engine=None, idx=qb1_id)
            cmd = Command(None, X, qubits=([qb0],), controls=[qb1])
            mapper._stored_commands = [cmd]
            mapper._send_possible_commands()
            assert len(mapper._stored_commands) == 1


@pytest.mark.parametrize("different_backend_ids", [False, True])
def test_send_possible_commands_allocate(different_backend_ids):
    if different_backend_ids:
        map_to_backend_ids = {0: 21, 1: 32, 2: 3, 3: 4, 4: 5, 5: 6}
    else:
        map_to_backend_ids = None
    mapper = two_d.GridMapper(num_rows=3, num_columns=2, mapped_ids_to_backend_ids=map_to_backend_ids)
    backend = DummyEngine(save_commands=True)
    backend.is_last_engine = True
    mapper.next_engine = backend
    qb0 = WeakQubitRef(engine=None, idx=0)
    cmd0 = Command(engine=None, gate=Allocate, qubits=([qb0],), controls=[], tags=[])
    mapper._stored_commands = [cmd0]
    mapper._currently_allocated_ids = set([10])
    # not in mapping:
    mapper.current_mapping = dict()
    assert len(backend.received_commands) == 0
    mapper._send_possible_commands()
    assert len(backend.received_commands) == 0
    assert mapper._stored_commands == [cmd0]
    # in mapping:
    mapper.current_mapping = {0: 3}
    mapper._send_possible_commands()
    assert len(mapper._stored_commands) == 0
    # Only self._run() sends Allocate gates
    mapped0 = WeakQubitRef(engine=None, idx=3)
    received_cmd = Command(
        engine=mapper,
        gate=Allocate,
        qubits=([mapped0],),
        controls=[],
        tags=[LogicalQubitIDTag(0)],
    )
    assert backend.received_commands[0] == received_cmd
    assert mapper._currently_allocated_ids == set([10, 0])


@pytest.mark.parametrize("different_backend_ids", [False, True])
def test_send_possible_commands_deallocate(different_backend_ids):
    if different_backend_ids:
        map_to_backend_ids = {0: 21, 1: 32, 2: 3, 3: 4, 4: 5, 5: 6}
    else:
        map_to_backend_ids = None
    mapper = two_d.GridMapper(num_rows=3, num_columns=2, mapped_ids_to_backend_ids=map_to_backend_ids)
    backend = DummyEngine(save_commands=True)
    backend.is_last_engine = True
    mapper.next_engine = backend
    qb0 = WeakQubitRef(engine=None, idx=0)
    cmd0 = Command(engine=None, gate=Deallocate, qubits=([qb0],), controls=[], tags=[])
    mapper._stored_commands = [cmd0]
    mapper.current_mapping = dict()
    mapper._currently_allocated_ids = set([10])
    # not yet allocated:
    mapper._send_possible_commands()
    assert len(backend.received_commands) == 0
    assert mapper._stored_commands == [cmd0]
    # allocated:
    mapper.current_mapping = {0: 3}
    mapper._currently_allocated_ids.add(0)
    mapper._send_possible_commands()
    assert len(backend.received_commands) == 1
    assert len(mapper._stored_commands) == 0
    assert mapper.current_mapping == dict()
    assert mapper._currently_allocated_ids == set([10])


@pytest.mark.parametrize("different_backend_ids", [False, True])
def test_send_possible_commands_keep_remaining_gates(different_backend_ids):
    if different_backend_ids:
        map_to_backend_ids = {0: 21, 1: 32, 2: 3, 3: 0, 4: 5, 5: 6}
    else:
        map_to_backend_ids = None
    mapper = two_d.GridMapper(num_rows=3, num_columns=2, mapped_ids_to_backend_ids=map_to_backend_ids)
    backend = DummyEngine(save_commands=True)
    backend.is_last_engine = True
    mapper.next_engine = backend
    qb0 = WeakQubitRef(engine=None, idx=0)
    qb1 = WeakQubitRef(engine=None, idx=1)
    cmd0 = Command(engine=None, gate=Allocate, qubits=([qb0],), controls=[], tags=[])
    cmd1 = Command(engine=None, gate=Deallocate, qubits=([qb0],), controls=[], tags=[])
    cmd2 = Command(engine=None, gate=Allocate, qubits=([qb1],), controls=[], tags=[])

    mapper._stored_commands = [cmd0, cmd1, cmd2]
    mapper.current_mapping = {0: 0}
    mapper._send_possible_commands()
    assert mapper._stored_commands == [cmd2]


@pytest.mark.parametrize("different_backend_ids", [False, True])
def test_send_possible_commands_one_inactive_qubit(different_backend_ids):
    if different_backend_ids:
        map_to_backend_ids = {0: 21, 1: 32, 2: 3, 3: 0, 4: 5, 5: 6}
    else:
        map_to_backend_ids = None
    mapper = two_d.GridMapper(num_rows=3, num_columns=2, mapped_ids_to_backend_ids=map_to_backend_ids)
    backend = DummyEngine(save_commands=True)
    backend.is_last_engine = True
    mapper.next_engine = backend
    qb0 = WeakQubitRef(engine=None, idx=0)
    qb1 = WeakQubitRef(engine=None, idx=1)
    cmd0 = Command(engine=None, gate=Allocate, qubits=([qb0],), controls=[], tags=[])
    cmd1 = Command(engine=None, gate=X, qubits=([qb0],), controls=[qb1])
    mapper._stored_commands = [cmd0, cmd1]
    mapper.current_mapping = {0: 0}
    mapper._send_possible_commands()
    assert mapper._stored_commands == [cmd1]


@pytest.mark.parametrize("different_backend_ids", [False, True])
@pytest.mark.parametrize("num_optimization_steps", [1, 10])
def test_run_and_receive(num_optimization_steps, different_backend_ids):
    if different_backend_ids:
        map_to_backend_ids = {0: 21, 1: 32, 2: 3, 3: 0}
    else:
        map_to_backend_ids = None

    def choose_last_permutation(swaps):
        choose_last_permutation.counter -= 1
        return choose_last_permutation.counter

    choose_last_permutation.counter = 100
    mapper = two_d.GridMapper(
        num_rows=2,
        num_columns=2,
        mapped_ids_to_backend_ids=map_to_backend_ids,
        optimization_function=choose_last_permutation,
        num_optimization_steps=num_optimization_steps,
    )
    backend = DummyEngine(save_commands=True)
    backend.is_last_engine = True
    mapper.next_engine = backend
    qb0 = WeakQubitRef(engine=None, idx=0)
    qb1 = WeakQubitRef(engine=None, idx=1)
    qb2 = WeakQubitRef(engine=None, idx=2)
    qb3 = WeakQubitRef(engine=None, idx=3)
    cmd0 = Command(engine=None, gate=Allocate, qubits=([qb0],))
    cmd1 = Command(engine=None, gate=Allocate, qubits=([qb1],))
    cmd2 = Command(engine=None, gate=Allocate, qubits=([qb2],))
    cmd3 = Command(engine=None, gate=Allocate, qubits=([qb3],))
    cmd4 = Command(None, X, qubits=([qb0],), controls=[qb1])
    cmd5 = Command(None, X, qubits=([qb1],), controls=[qb3])
    cmd6 = Command(None, X, qubits=([qb3],), controls=[qb2])
    cmd7 = Command(None, X, qubits=([qb0],), controls=[qb2])
    cmd8 = Command(engine=None, gate=Deallocate, qubits=([qb1],))
    all_cmd = [cmd0, cmd1, cmd2, cmd3, cmd4, cmd5, cmd6, cmd7, cmd8]
    mapper.receive(all_cmd)
    assert mapper._stored_commands == all_cmd
    qb4 = WeakQubitRef(engine=None, idx=-1)
    cmd_flush = Command(engine=None, gate=FlushGate(), qubits=([qb4],))
    mapper.receive([cmd_flush])
    assert mapper._stored_commands == []
    assert len(backend.received_commands) == 10
    assert mapper._currently_allocated_ids == set([0, 2, 3])
    if different_backend_ids:
        assert (
            mapper.current_mapping == {0: 21, 2: 3, 3: 0}
            or mapper.current_mapping == {0: 32, 2: 0, 3: 21}
            or mapper.current_mapping == {0: 3, 2: 21, 3: 32}
            or mapper.current_mapping == {0: 0, 2: 32, 3: 3}
        )
    else:
        assert (
            mapper.current_mapping == {0: 0, 2: 2, 3: 3}
            or mapper.current_mapping == {0: 1, 2: 3, 3: 0}
            or mapper.current_mapping == {0: 2, 2: 0, 3: 1}
            or mapper.current_mapping == {0: 3, 2: 1, 3: 2}
        )
    cmd9 = Command(None, X, qubits=([qb0],), controls=[qb3])
    mapper.storage = 1
    mapper.receive([cmd9])
    assert mapper._currently_allocated_ids == set([0, 2, 3])
    assert mapper._stored_commands == []
    assert len(mapper.current_mapping) == 3
    assert 0 in mapper.current_mapping
    assert 2 in mapper.current_mapping
    assert 3 in mapper.current_mapping
    assert mapper.num_mappings == 1


def test_run_infinite_loop_detection():
    mapper = two_d.GridMapper(num_rows=2, num_columns=2)
    backend = DummyEngine(save_commands=True)
    backend.is_last_engine = True
    mapper.next_engine = backend
    qb0 = WeakQubitRef(engine=None, idx=0)
    qb1 = WeakQubitRef(engine=None, idx=1)
    qb2 = WeakQubitRef(engine=None, idx=2)
    qb3 = WeakQubitRef(engine=None, idx=3)
    qb4 = WeakQubitRef(engine=None, idx=4)
    cmd0 = Command(engine=None, gate=Allocate, qubits=([qb0],))
    cmd1 = Command(engine=None, gate=Allocate, qubits=([qb1],))
    cmd2 = Command(engine=None, gate=Allocate, qubits=([qb2],))
    cmd3 = Command(engine=None, gate=Allocate, qubits=([qb3],))
    cmd4 = Command(engine=None, gate=Allocate, qubits=([qb4],))
    cmd5 = Command(None, X, qubits=([qb0],), controls=[qb1])
    qb2 = WeakQubitRef(engine=None, idx=-1)
    cmd_flush = Command(engine=None, gate=FlushGate(), qubits=([qb2],))
    with pytest.raises(RuntimeError):
        mapper.receive([cmd0, cmd1, cmd2, cmd3, cmd4, cmd5, cmd_flush])


def test_correct_stats():
    # Should test stats for twice same mapping but depends on heuristic
    mapper = two_d.GridMapper(num_rows=3, num_columns=1)
    backend = DummyEngine(save_commands=True)
    backend.is_last_engine = True
    mapper.next_engine = backend
    qb0 = WeakQubitRef(engine=None, idx=0)
    qb1 = WeakQubitRef(engine=None, idx=1)
    qb2 = WeakQubitRef(engine=None, idx=2)
    cmd0 = Command(engine=None, gate=Allocate, qubits=([qb0],))
    cmd1 = Command(engine=None, gate=Allocate, qubits=([qb1],))
    cmd2 = Command(engine=None, gate=Allocate, qubits=([qb2],))
    cmd3 = Command(None, X, qubits=([qb0],), controls=[qb1])
    cmd4 = Command(None, X, qubits=([qb1],), controls=[qb2])
    cmd5 = Command(None, X, qubits=([qb0],), controls=[qb2])
    cmd6 = Command(None, X, qubits=([qb2],), controls=[qb1])
    cmd7 = Command(None, X, qubits=([qb0],), controls=[qb1])
    cmd8 = Command(None, X, qubits=([qb1],), controls=[qb2])
    qb_flush = WeakQubitRef(engine=None, idx=-1)
    cmd_flush = Command(engine=None, gate=FlushGate(), qubits=([qb_flush],))
    mapper.receive([cmd0, cmd1, cmd2, cmd3, cmd4, cmd5, cmd6, cmd7, cmd8, cmd_flush])
    assert mapper.num_mappings == 2


def test_send_possible_cmds_before_new_mapping():
    mapper = two_d.GridMapper(num_rows=3, num_columns=1)
    backend = DummyEngine(save_commands=True)
    backend.is_last_engine = True
    mapper.next_engine = backend

    def dont_call_mapping():
        raise Exception

    mapper._return_new_mapping = dont_call_mapping
    mapper.current_mapping = {0: 1}
    qb0 = WeakQubitRef(engine=None, idx=0)
    cmd0 = Command(engine=None, gate=Allocate, qubits=([qb0],))
    qb2 = WeakQubitRef(engine=None, idx=-1)
    cmd_flush = Command(engine=None, gate=FlushGate(), qubits=([qb2],))
    mapper.receive([cmd0, cmd_flush])


def test_logical_id_tags_allocate_and_deallocate():
    mapper = two_d.GridMapper(num_rows=2, num_columns=2)
    backend = DummyEngine(save_commands=True)
    backend.is_last_engine = True
    mapper.next_engine = backend
    qb0 = WeakQubitRef(engine=None, idx=0)
    qb1 = WeakQubitRef(engine=None, idx=1)
    cmd0 = Command(engine=None, gate=Allocate, qubits=([qb0],))
    cmd1 = Command(engine=None, gate=Allocate, qubits=([qb1],))
    cmd2 = Command(None, X, qubits=([qb0],), controls=[qb1])
    cmd3 = Command(engine=None, gate=Deallocate, qubits=([qb0],))
    cmd4 = Command(engine=None, gate=Deallocate, qubits=([qb1],))
    mapper.current_mapping = {0: 0, 1: 3}
    qb_flush = WeakQubitRef(engine=None, idx=-1)
    cmd_flush = Command(engine=None, gate=FlushGate(), qubits=([qb_flush],))
    mapper.receive([cmd0, cmd1, cmd2, cmd_flush])
    assert backend.received_commands[0].gate == Allocate
    assert backend.received_commands[0].qubits[0][0].id == 0
    assert backend.received_commands[0].tags == [LogicalQubitIDTag(0)]
    assert backend.received_commands[1].gate == Allocate
    assert backend.received_commands[1].qubits[0][0].id == 3
    assert backend.received_commands[1].tags == [LogicalQubitIDTag(1)]
    for cmd in backend.received_commands[2:]:
        if cmd.gate == Allocate:
            assert cmd.tags == []
        elif cmd.gate == Deallocate:
            assert cmd.tags == []
    mapped_id_for_0 = mapper.current_mapping[0]
    mapped_id_for_1 = mapper.current_mapping[1]
    mapper.receive([cmd3, cmd4, cmd_flush])
    assert backend.received_commands[-3].gate == Deallocate
    assert backend.received_commands[-3].qubits[0][0].id == mapped_id_for_0
    assert backend.received_commands[-3].tags == [LogicalQubitIDTag(0)]
    assert backend.received_commands[-2].gate == Deallocate
    assert backend.received_commands[-2].qubits[0][0].id == mapped_id_for_1
    assert backend.received_commands[-2].tags == [LogicalQubitIDTag(1)]


def test_check_that_local_optimizer_doesnt_merge():
    mapper = two_d.GridMapper(num_rows=2, num_columns=2)
    optimizer = LocalOptimizer(10)
    backend = DummyEngine(save_commands=True)
    backend.is_last_engine = True
    mapper.next_engine = optimizer
    optimizer.next_engine = backend
    mapper.current_mapping = {0: 0}
    mapper.storage = 1
    qb0 = WeakQubitRef(engine=None, idx=0)
    qb1 = WeakQubitRef(engine=None, idx=1)
    qb_flush = WeakQubitRef(engine=None, idx=-1)
    cmd_flush = Command(engine=None, gate=FlushGate(), qubits=([qb_flush],))
    cmd0 = Command(engine=None, gate=Allocate, qubits=([qb0],))
    cmd1 = Command(None, X, qubits=([qb0],))
    cmd2 = Command(engine=None, gate=Deallocate, qubits=([qb0],))
    mapper.receive([cmd0, cmd1, cmd2])
    assert len(mapper._stored_commands) == 0
    mapper.current_mapping = {1: 0}
    cmd3 = Command(engine=None, gate=Allocate, qubits=([qb1],))
    cmd4 = Command(None, X, qubits=([qb1],))
    cmd5 = Command(engine=None, gate=Deallocate, qubits=([qb1],))
    mapper.receive([cmd3, cmd4, cmd5, cmd_flush])
    assert len(backend.received_commands) == 7



# -*- coding: utf-8 -*-
#   Copyright 2018 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import cmath

import pytest

from projectq import MainEngine
from projectq.backends import Simulator
from projectq.cengines import (
    AutoReplacer,
    DecompositionRuleSet,
    DummyEngine,
    InstructionFilter,
)
from projectq.meta import Control
from projectq.ops import All, Measure, Ph, QubitOperator, X, Y, Z, Command
from projectq.types import WeakQubitRef

import projectq.setups.decompositions.qubitop2onequbit as qubitop2onequbit


def test_recognize():
    saving_backend = DummyEngine(save_commands=True)
    eng = MainEngine(backend=saving_backend, engine_list=[])
    ctrl_qureg = eng.allocate_qureg(2)
    qureg = eng.allocate_qureg(2)
    with Control(eng, ctrl_qureg):
        QubitOperator("X0 Y1") | qureg
    with Control(eng, ctrl_qureg[0]):
        QubitOperator("X0 Y1") | qureg
    eng.flush()
    cmd0 = saving_backend.received_commands[4]
    cmd1 = saving_backend.received_commands[5]
    assert not qubitop2onequbit._recognize_qubitop(cmd0)
    assert qubitop2onequbit._recognize_qubitop(cmd1)


def _decomp_gates(eng, cmd):
    if isinstance(cmd.gate, QubitOperator):
        return False
    else:
        return True


def test_qubitop2singlequbit_invalid():
    qb0 = WeakQubitRef(None, idx=0)
    qb1 = WeakQubitRef(None, idx=1)
    with pytest.raises(ValueError):
        qubitop2onequbit._decompose_qubitop(Command(None, QubitOperator(), ([qb0], [qb1])))


def test_qubitop2singlequbit():
    num_qubits = 4
    random_initial_state = [0.2 + 0.1 * x * cmath.exp(0.1j + 0.2j * x) for x in range(2 ** (num_qubits + 1))]
    rule_set = DecompositionRuleSet(modules=[qubitop2onequbit])
    test_eng = MainEngine(
        backend=Simulator(),
        engine_list=[AutoReplacer(rule_set), InstructionFilter(_decomp_gates)],
    )
    test_qureg = test_eng.allocate_qureg(num_qubits)
    test_ctrl_qb = test_eng.allocate_qubit()
    test_eng.flush()
    test_eng.backend.set_wavefunction(random_initial_state, test_qureg + test_ctrl_qb)
    correct_eng = MainEngine()
    correct_qureg = correct_eng.allocate_qureg(num_qubits)
    correct_ctrl_qb = correct_eng.allocate_qubit()
    correct_eng.flush()
    correct_eng.backend.set_wavefunction(random_initial_state, correct_qureg + correct_ctrl_qb)

    qubit_op_0 = QubitOperator("X0 Y1 Z3", -1.0j)
    qubit_op_1 = QubitOperator("Z0 Y1 X3", cmath.exp(0.6j))

    qubit_op_0 | test_qureg
    with Control(test_eng, test_ctrl_qb):
        qubit_op_1 | test_qureg
    test_eng.flush()

    correct_eng.backend.apply_qubit_operator(qubit_op_0, correct_qureg)
    with Control(correct_eng, correct_ctrl_qb):
        Ph(0.6) | correct_qureg[0]
        Z | correct_qureg[0]
        Y | correct_qureg[1]
        X | correct_qureg[3]
    correct_eng.flush()

    for fstate in range(2 ** (num_qubits + 1)):
        binary_state = format(fstate, '0' + str(num_qubits + 1) + 'b')
        test = test_eng.backend.get_amplitude(binary_state, test_qureg + test_ctrl_qb)
        correct = correct_eng.backend.get_amplitude(binary_state, correct_qureg + correct_ctrl_qb)
        assert correct == pytest.approx(test, rel=1e-10, abs=1e-10)

    All(Measure) | correct_qureg + correct_ctrl_qb
    All(Measure) | test_qureg + test_ctrl_qb
    correct_eng.flush()
    test_eng.flush()



# -*- coding: utf-8 -*-
# pylint: skip-file

import math

from projectq import MainEngine
from projectq.ops import H, Z, X, Measure, All
from projectq.meta import Loop, Compute, Uncompute, Control


def run_grover(eng, n, oracle):
    """
    Runs Grover's algorithm on n qubit using the provided quantum oracle.

    Args:
        eng (MainEngine): Main compiler engine to run Grover on.
        n (int): Number of bits in the solution.
        oracle (function): Function accepting the engine, an n-qubit register,
            and an output qubit which is flipped by the oracle for the correct
            bit string.

    Returns:
        solution (list<int>): Solution bit-string.
    """
    x = eng.allocate_qureg(n)

    # start in uniform superposition
    All(H) | x

    # number of iterations we have to run:
    num_it = int(math.pi / 4.0 * math.sqrt(1 << n))

    # prepare the oracle output qubit (the one that is flipped to indicate the
    # solution. start in state 1/sqrt(2) * (|0> - |1>) s.t. a bit-flip turns
    # into a (-1)-phase.
    oracle_out = eng.allocate_qubit()
    X | oracle_out
    H | oracle_out

    # run num_it iterations
    with Loop(eng, num_it):
        # oracle adds a (-1)-phase to the solution
        oracle(eng, x, oracle_out)

        # reflection across uniform superposition
        with Compute(eng):
            All(H) | x
            All(X) | x

        with Control(eng, x[0:-1]):
            Z | x[-1]

        Uncompute(eng)

    All(Measure) | x
    Measure | oracle_out

    eng.flush()
    # return result
    return [int(qubit) for qubit in x]


def alternating_bits_oracle(eng, qubits, output):
    """
    Marks the solution string 1,0,1,0,...,0,1 by flipping the output qubit,
    conditioned on qubits being equal to the alternating bit-string.

    Args:
        eng (MainEngine): Main compiler engine the algorithm is being run on.
        qubits (Qureg): n-qubit quantum register Grover search is run on.
        output (Qubit): Output qubit to flip in order to mark the solution.
    """
    with Compute(eng):
        All(X) | qubits[1::2]
    with Control(eng, qubits):
        X | output
    Uncompute(eng)


if __name__ == "__main__":
    eng = MainEngine()  # use default compiler engine
    # run Grover search to find a 7-bit solution
    print(run_grover(eng, 7, alternating_bits_oracle))



# -*- coding: utf-8 -*-
#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"Tests for projectq.setups.decompositions.rx2rz.py"

import math

import pytest

from projectq.backends import Simulator
from projectq.cengines import (
    AutoReplacer,
    DecompositionRuleSet,
    DummyEngine,
    InstructionFilter,
    MainEngine,
)
from projectq.meta import Control
from projectq.ops import Measure, Rx

from . import rx2rz


def test_recognize_correct_gates():
    saving_backend = DummyEngine(save_commands=True)
    eng = MainEngine(backend=saving_backend)
    qubit = eng.allocate_qubit()
    ctrl_qubit = eng.allocate_qubit()
    eng.flush()
    Rx(0.3) | qubit
    with Control(eng, ctrl_qubit):
        Rx(0.4) | qubit
    eng.flush(deallocate_qubits=True)
    assert rx2rz._recognize_RxNoCtrl(saving_backend.received_commands[3])
    assert not rx2rz._recognize_RxNoCtrl(saving_backend.received_commands[4])


def rx_decomp_gates(eng, cmd):
    g = cmd.gate
    if isinstance(g, Rx):
        return False
    else:
        return True


@pytest.mark.parametrize("angle", [0, math.pi, 2 * math.pi, 4 * math.pi, 0.5])
def test_decomposition(angle):
    for basis_state in ([1, 0], [0, 1]):
        correct_dummy_eng = DummyEngine(save_commands=True)
        correct_eng = MainEngine(backend=Simulator(), engine_list=[correct_dummy_eng])

        rule_set = DecompositionRuleSet(modules=[rx2rz])
        test_dummy_eng = DummyEngine(save_commands=True)
        test_eng = MainEngine(
            backend=Simulator(),
            engine_list=[
                AutoReplacer(rule_set),
                InstructionFilter(rx_decomp_gates),
                test_dummy_eng,
            ],
        )

        correct_qb = correct_eng.allocate_qubit()
        Rx(angle) | correct_qb
        correct_eng.flush()

        test_qb = test_eng.allocate_qubit()
        Rx(angle) | test_qb
        test_eng.flush()

        assert correct_dummy_eng.received_commands[1].gate == Rx(angle)
        assert test_dummy_eng.received_commands[1].gate != Rx(angle)

        for fstate in ['0', '1']:
            test = test_eng.backend.get_amplitude(fstate, test_qb)
            correct = correct_eng.backend.get_amplitude(fstate, correct_qb)
            assert correct == pytest.approx(test, rel=1e-12, abs=1e-12)

        Measure | test_qb
        Measure | correct_qb



# -*- coding: utf-8 -*-
#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for projectq.ops._gates."""

import cmath
import math
import numpy as np
import pytest

from projectq.types import Qubit
from projectq import MainEngine
from projectq.cengines import DummyEngine
from projectq.ops import (
    T,
    Y,
    NotInvertible,
    Entangle,
    Rx,
    FastForwardingGate,
    Command,
    C,
    ClassicalInstructionGate,
    All,
)
from projectq.types import WeakQubitRef

from projectq.ops import _metagates


def test_tensored_gate_invalid():
    qb0 = WeakQubitRef(None, idx=0)
    qb1 = WeakQubitRef(None, idx=1)

    with pytest.raises(ValueError):
        _metagates.Tensor(Y) | (qb0, qb1)

    with pytest.raises(ValueError):
        _metagates.Tensor(Y) | qb0


def test_tensored_controlled_gate():
    saving_backend = DummyEngine(save_commands=True)
    main_engine = MainEngine(backend=saving_backend, engine_list=[DummyEngine()])
    gate = Rx(0.6)
    qubit0 = Qubit(main_engine, 0)
    qubit1 = Qubit(main_engine, 1)
    qubit2 = Qubit(main_engine, 2)
    target_qubits = [qubit1, qubit2]
    C(All(gate)) | (qubit0, target_qubits)

    assert saving_backend.received_commands[-1].gate == gate
    assert len(saving_backend.received_commands[-1].control_qubits) == 1


def test_daggered_gate_init():
    # Choose gate which does not have an inverse gate:
    not_invertible_gate = T
    with pytest.raises(NotInvertible):
        not_invertible_gate.get_inverse()
    # Choose gate which does have an inverse defined:
    invertible_gate = Y
    assert invertible_gate.get_inverse() == Y
    # Test init and matrix
    dagger_inv = _metagates.DaggeredGate(not_invertible_gate)
    assert dagger_inv._gate == not_invertible_gate
    assert np.array_equal(dagger_inv.matrix, np.matrix([[1, 0], [0, cmath.exp(-1j * cmath.pi / 4)]]))
    inv = _metagates.DaggeredGate(invertible_gate)
    assert inv._gate == invertible_gate
    assert np.array_equal(inv.matrix, np.matrix([[0, -1j], [1j, 0]]))
    # Test matrix
    no_matrix_gate = Entangle
    with pytest.raises(AttributeError):
        no_matrix_gate.matrix
    inv_no_matrix_gate = _metagates.DaggeredGate(no_matrix_gate)
    with pytest.raises(AttributeError):
        inv_no_matrix_gate.matrix


def test_daggered_gate_str():
    daggered_gate = _metagates.DaggeredGate(Y)
    assert str(daggered_gate) == str(Y) + r"^\dagger"


def test_daggered_gate_hashable():
    daggered_gate1 = _metagates.DaggeredGate(Y)
    daggered_gate2 = _metagates.DaggeredGate(T)
    d = {daggered_gate1: 1, daggered_gate2: 3}
    assert len(d) == 2
    # for efficiency reasons the following should be true:
    assert hash(daggered_gate1) != hash(daggered_gate2)


def test_daggered_gate_tex_str():
    daggered_gate = _metagates.DaggeredGate(Y)
    str_Y = Y.tex_str() if hasattr(Y, 'tex_str') else str(Y)
    assert daggered_gate.tex_str() == str_Y + r"${}^\dagger$"

    # test for a gate with tex_str method
    rx = Rx(0.5)
    daggered_rx = _metagates.DaggeredGate(rx)
    str_rx = rx.tex_str() if hasattr(rx, 'tex_str') else str(rx)
    assert daggered_rx.tex_str() == str_rx + r"${}^\dagger$"


def test_daggered_gate_get_inverse():
    daggered_gate = _metagates.DaggeredGate(Y)
    assert daggered_gate.get_inverse() == Y


def test_daggered_gate_comparison():
    daggered_gate = _metagates.DaggeredGate(Y)
    daggered_gate2 = _metagates.DaggeredGate(Y)
    assert daggered_gate == daggered_gate2


def test_get_inverse():
    # Choose gate which does not have an inverse gate:
    not_invertible_gate = T
    with pytest.raises(NotInvertible):
        not_invertible_gate.get_inverse()
    # Choose gate which does have an inverse defined:
    invertible_gate = Y
    assert invertible_gate.get_inverse() == Y
    # Check get_inverse(gate)
    inv = _metagates.get_inverse(not_invertible_gate)
    assert isinstance(inv, _metagates.DaggeredGate) and inv._gate == not_invertible_gate
    inv2 = _metagates.get_inverse(invertible_gate)
    assert inv2 == Y


def test_is_identity():
    # Choose gate which is not an identity gate:
    non_identity_gate = Rx(0.5)
    assert not non_identity_gate.is_identity()
    assert not _metagates.is_identity(non_identity_gate)
    # Choose gate which is an identity gate:
    identity_gate = Rx(0.0)
    assert identity_gate.is_identity()
    assert _metagates.is_identity(identity_gate)


def test_controlled_gate_init():
    one_control = _metagates.ControlledGate(Y, 1)
    two_control = _metagates.ControlledGate(Y, 2)
    three_control = _metagates.ControlledGate(one_control, 2)
    assert one_control._gate == Y
    assert one_control._n == 1
    assert two_control._gate == Y
    assert two_control._n == 2
    assert three_control._gate == Y
    assert three_control._n == 3


def test_controlled_gate_str():
    one_control = _metagates.ControlledGate(Y, 2)
    assert str(one_control) == "CC" + str(Y)


def test_controlled_gate_get_inverse():
    one_control = _metagates.ControlledGate(Rx(0.5), 1)
    expected = _metagates.ControlledGate(Rx(-0.5 + 4 * math.pi), 1)
    assert one_control.get_inverse() == expected


def test_controlled_gate_empty_controls():
    rec = DummyEngine(save_commands=True)
    eng = MainEngine(backend=rec, engine_list=[])

    a = eng.allocate_qureg(1)
    _metagates.ControlledGate(Y, 0) | ((), a)
    assert rec.received_commands[-1] == Command(eng, Y, [a])


def test_controlled_gate_or():
    saving_backend = DummyEngine(save_commands=True)
    main_engine = MainEngine(backend=saving_backend, engine_list=[DummyEngine()])
    gate = Rx(0.6)
    qubit0 = Qubit(main_engine, 0)
    qubit1 = Qubit(main_engine, 1)
    qubit2 = Qubit(main_engine, 2)
    qubit3 = Qubit(main_engine, 3)
    expected_cmd = Command(main_engine, gate, ([qubit3],), controls=[qubit0, qubit1, qubit2])
    received_commands = []
    # Option 1:
    _metagates.ControlledGate(gate, 3) | ([qubit1], [qubit0], [qubit2], [qubit3])
    # Option 2:
    _metagates.ControlledGate(gate, 3) | (qubit1, qubit0, qubit2, qubit3)
    # Option 3:
    _metagates.ControlledGate(gate, 3) | ([qubit1, qubit0], qubit2, qubit3)
    # Option 4:
    _metagates.ControlledGate(gate, 3) | (qubit1, [qubit0, qubit2], qubit3)
    # Wrong option 5:
    with pytest.raises(_metagates.ControlQubitError):
        _metagates.ControlledGate(gate, 3) | (qubit1, [qubit0, qubit2, qubit3])
    # Remove Allocate and Deallocate gates
    for cmd in saving_backend.received_commands:
        if not (isinstance(cmd.gate, FastForwardingGate) or isinstance(cmd.gate, ClassicalInstructionGate)):
            received_commands.append(cmd)
    assert len(received_commands) == 4
    for cmd in received_commands:
        assert cmd == expected_cmd


def test_controlled_gate_comparison():
    gate1 = _metagates.ControlledGate(Y, 1)
    gate2 = _metagates.ControlledGate(Y, 1)
    gate3 = _metagates.ControlledGate(T, 1)
    gate4 = _metagates.ControlledGate(Y, 2)
    assert gate1 == gate2
    assert not gate1 == gate3
    assert gate1 != gate4


def test_c():
    expected = _metagates.ControlledGate(Y, 2)
    assert _metagates.C(Y, 2) == expected


def test_tensor_init():
    gate = _metagates.Tensor(Y)
    assert gate._gate == Y


def test_tensor_str():
    gate = _metagates.Tensor(Y)
    assert str(gate) == "Tensor(" + str(Y) + ")"


def test_tensor_get_inverse():
    gate = _metagates.Tensor(Rx(0.6))
    inverse = gate.get_inverse()
    assert isinstance(inverse, _metagates.Tensor)
    assert inverse._gate == Rx(-0.6 + 4 * math.pi)


def test_tensor_comparison():
    gate1 = _metagates.Tensor(Rx(0.6))
    gate2 = _metagates.Tensor(Rx(0.6 + 4 * math.pi))
    assert gate1 == gate2
    assert gate1 != Rx(0.6)


def test_tensor_or():
    saving_backend = DummyEngine(save_commands=True)
    main_engine = MainEngine(backend=saving_backend, engine_list=[DummyEngine()])
    gate = Rx(0.6)
    qubit0 = Qubit(main_engine, 0)
    qubit1 = Qubit(main_engine, 1)
    qubit2 = Qubit(main_engine, 2)
    # Option 1:
    _metagates.Tensor(gate) | ([qubit0, qubit1, qubit2],)
    # Option 2:
    _metagates.Tensor(gate) | [qubit0, qubit1, qubit2]
    received_commands = []
    # Remove Allocate and Deallocate gates
    for cmd in saving_backend.received_commands:
        if not (isinstance(cmd.gate, FastForwardingGate) or isinstance(cmd.gate, ClassicalInstructionGate)):
            received_commands.append(cmd)
    # Check results
    assert len(received_commands) == 6
    qubit_ids = []
    for cmd in received_commands:
        assert len(cmd.qubits) == 1
        assert cmd.gate == gate
        qubit_ids.append(cmd.qubits[0][0].id)
    assert sorted(qubit_ids) == [0, 0, 1, 1, 2, 2]



# -*- coding: utf-8 -*-
#   Copyright 2020 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pytest
import matplotlib
import matplotlib.pyplot as plt  # noqa: F401

from projectq import MainEngine
from projectq.ops import H, C, X, Measure, All, AllocateQubitGate, FlushGate
from projectq.cengines import DummyEngine, BasicEngine
from projectq.backends import Simulator
from projectq.libs.hist import histogram


@pytest.fixture(scope="module")
def matplotlib_setup():
    old_backend = matplotlib.get_backend()
    matplotlib.use('agg')  # avoid showing the histogram plots
    yield
    matplotlib.use(old_backend)


def test_invalid_backend(matplotlib_setup):
    eng = MainEngine(backend=DummyEngine())
    qubit = eng.allocate_qubit()
    eng.flush()

    with pytest.raises(RuntimeError):
        histogram(eng.backend, qubit)


def test_backend_get_probabilities_method(matplotlib_setup):
    class MyBackend(BasicEngine):
        def get_probabilities(self, qureg):
            return {'000': 0.5, '111': 0.5}

        def is_available(self, cmd):
            return True

        def receive(self, command_list):
            for cmd in command_list:
                if not isinstance(cmd.gate, FlushGate):
                    assert isinstance(cmd.gate, AllocateQubitGate)

    eng = MainEngine(backend=MyBackend(), verbose=True)
    qureg = eng.allocate_qureg(3)
    eng.flush()
    _, _, prob = histogram(eng.backend, qureg)
    assert prob['000'] == 0.5
    assert prob['111'] == 0.5

    # NB: avoid throwing exceptions when destroying the MainEngine
    eng.next_engine = DummyEngine()
    eng.next_engine.is_last_engine = True


def test_qubit(matplotlib_setup):
    sim = Simulator()
    eng = MainEngine(sim)
    qubit = eng.allocate_qubit()
    eng.flush()
    _, _, prob = histogram(sim, qubit)
    assert prob["0"] == pytest.approx(1)
    assert prob["1"] == pytest.approx(0)
    H | qubit
    eng.flush()
    _, _, prob = histogram(sim, qubit)
    assert prob["0"] == pytest.approx(0.5)
    Measure | qubit
    eng.flush()
    _, _, prob = histogram(sim, qubit)
    assert prob["0"] == pytest.approx(1) or prob["1"] == pytest.approx(1)


def test_qureg(matplotlib_setup):
    sim = Simulator()
    eng = MainEngine(sim)
    qureg = eng.allocate_qureg(3)
    eng.flush()
    _, _, prob = histogram(sim, qureg)
    assert prob["000"] == pytest.approx(1)
    assert prob["110"] == pytest.approx(0)
    H | qureg[0]
    C(X, 1) | (qureg[0], qureg[1])
    H | qureg[2]
    eng.flush()
    _, _, prob = histogram(sim, qureg)
    assert prob["110"] == pytest.approx(0.25)
    assert prob["100"] == pytest.approx(0)
    All(Measure) | qureg
    eng.flush()
    _, _, prob = histogram(sim, qureg)
    assert (
        prob["000"] == pytest.approx(1)
        or prob["001"] == pytest.approx(1)
        or prob["110"] == pytest.approx(1)
        or prob["111"] == pytest.approx(1)
    )
    assert prob["000"] + prob["001"] + prob["110"] + prob["111"] == pytest.approx(1)


def test_combination(matplotlib_setup):
    sim = Simulator()
    eng = MainEngine(sim)
    qureg = eng.allocate_qureg(2)
    qubit = eng.allocate_qubit()
    eng.flush()
    _, _, prob = histogram(sim, [qureg, qubit])
    assert prob["000"] == pytest.approx(1)
    H | qureg[0]
    C(X, 1) | (qureg[0], qureg[1])
    H | qubit
    Measure | qureg[0]
    eng.flush()
    _, _, prob = histogram(sim, [qureg, qubit])
    assert (prob["000"] == pytest.approx(0.5) and prob["001"] == pytest.approx(0.5)) or (
        prob["110"] == pytest.approx(0.5) and prob["111"] == pytest.approx(0.5)
    )
    assert prob["100"] == pytest.approx(0)
    Measure | qubit


def test_too_many_qubits(matplotlib_setup, capsys):
    sim = Simulator()
    eng = MainEngine(sim)
    qureg = eng.allocate_qureg(6)
    eng.flush()
    l_ref = len(capsys.readouterr().out)
    _, _, prob = histogram(sim, qureg)
    assert len(capsys.readouterr().out) > l_ref
    assert prob["000000"] == pytest.approx(1)
    All(Measure)
